package il.co.iter.dbdom;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;

import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.Attr;
import org.w3c.dom.DOMException;

/**
 * Objects implementing the <code>NamedNodeMap</code> interface are used to 
 * represent collections of nodes that can be accessed by name. Note that 
 * <code>NamedNodeMap</code> does not inherit from <code>NodeList</code>; 
 * <code>NamedNodeMaps</code> are not maintained in any particular order. 
 * Objects contained in an object implementing <code>NamedNodeMap</code> may 
 * also be accessed by an ordinal index, but this is simply to allow 
 * convenient enumeration of the contents of a <code>NamedNodeMap</code>, 
 * and does not imply that the DOM specifies an order to these Nodes. 
 * <p><code>NamedNodeMap</code> objects in the DOM are live.
 * <p>See also the <a href='http://www.w3.org/TR/2000/REC-DOM-Level-2-Core-20001113'>Document Object Model (DOM) Level 2 Core Specification</a>.
 *
 * <p>This class represents a collection of {@link DBAttr}s associated with a
 * {@link DBElement}.
 *
 * <p>Implementation of this class is unusual for this package it that some of
 * its methods call Java methods on an associated {@link DBElement}, instead
 * of stored procedures in the database. this is because
 * {@link org.w3c.dom.Element} provides these methods, and because at current,
 * there are no correponding SPs in the database. This situation may change in
 * the future.
 */
public class DBAttrNodeMap implements NamedNodeMap {

    private DBElement element;
    private long elementId;
    private DBDOMImplementation dbDomImplementation;
    private Connection con;

    /**
     * Constructs a {@link DBAttrNodeMap} associated with a {@link DBElement}
     */
    protected DBAttrNodeMap (DBElement element) {
        this.element = element;
        this.elementId = element.getId();
        this.dbDomImplementation = element.getDbDomImplementation();
        this.con = dbDomImplementation.getConnection();
    }

    /**
     * Retrieves a node specified by name.
     * @param nameThe <code>nodeName</code> of a node to retrieve.
     * @return A <code>Node</code> (of any type) with the specified 
     *   <code>nodeName</code>, or <code>null</code> if it does not identify 
     *   any node in this map.
     *
     * <p>Calls {@link DBElement#getAttributeNode(String)}
     */
    public Node getNamedItem(String name) {
        return element.getAttributeNode (name);
    }

    /**
     * Adds a node using its <code>nodeName</code> attribute. If a node with 
     * that name is already present in this map, it is replaced by the new 
     * one.
     * <br>As the <code>nodeName</code> attribute is used to derive the name 
     * which the node must be stored under, multiple nodes of certain types 
     * (those that have a "special" string value) cannot be stored as the 
     * names would clash. This is seen as preferable to allowing nodes to be 
     * aliased.
     * @param argA node to store in this map. The node will later be 
     *   accessible using the value of its <code>nodeName</code> attribute.
     * @return If the new <code>Node</code> replaces an existing node the 
     *   replaced <code>Node</code> is returned, otherwise <code>null</code> 
     *   is returned.
     * @exception DOMException
     *   WRONG_DOCUMENT_ERR: Raised if <code>arg</code> was created from a 
     *   different document than the one that created this map.
     *   <br>NO_MODIFICATION_ALLOWED_ERR: Raised if this map is readonly.
     *   <br>INUSE_ATTRIBUTE_ERR: Raised if <code>arg</code> is an
     *   <code>Attr</code> that is already an attribute of another 
     *   <code>Element</code> object. The DOM user must explicitly clone 
     *   <code>Attr</code> nodes to re-use them in other elements.
     *
     * <p>Calls {@link DBElement#setAttributeNode(Attr)}
     */
    public Node setNamedItem(Node arg)
                             throws DOMException {
        try {
            return element.setAttributeNode ((Attr)arg);
        } catch (ClassCastException cce) {
            throw new DOMException (DOMException.HIERARCHY_REQUEST_ERR, "Only attributes are allowed in this Map");
        }
    }

    /** <string>Not Implemented</strong>
     * Removes a node specified by name. When this map contains the attributes 
     * attached to an element, if the removed attribute is known to have a 
     * default value, an attribute immediately appears containing the 
     * default value as well as the corresponding namespace URI, local name, 
     * and prefix when applicable.
     * @param nameThe <code>nodeName</code> of the node to remove.
     * @return The node removed from this map if a node with such a name 
     *   exists.
     * @exception DOMException
     *   NOT_FOUND_ERR: Raised if there is no node named <code>name</code> in 
     *   this map.
     *   <br>NO_MODIFICATION_ALLOWED_ERR: Raised if this map is readonly.
     */
    public Node removeNamedItem(String name)
                                throws DOMException {
        throw new DOMException(DOMException.NOT_SUPPORTED_ERR, "this feature is not implemented yet");
    }

    /** <strong>Not Tested</strong>
     * Returns the <code>index</code>th item in the map. If <code>index</code> 
     * is greater than or equal to the number of nodes in this map, this 
     * returns <code>null</code>.
     * @param indexIndex into this map.
     * @return The node at the <code>index</code>th position in the map, or 
     *   <code>null</code> if that is not a valid index.
     */
    public Node item(int index) {
        try {
            String sqlString = "select namedMapItem (?, ?, ?)";
            PreparedStatement statement = (PreparedStatement)dbDomImplementation.getStatements().get(sqlString);
            // if statement is null now, it either isn't present in statements,
            // or is present but uninitialized, which is highly unlikely.
            // putting it into the map wil address both cases
            if (statement == null) {
                statement = con.prepareStatement(sqlString);
                dbDomImplementation.getStatements().put(sqlString, statement);
            }
            statement.setLong (1, elementId);
            statement.setInt  (2, Node.ATTRIBUTE_NODE);
            statement.setInt  (3, index);
            ResultSet resultSet = statement.executeQuery();
            if (!resultSet.next()) {
                throw new DBDOMSpException (sqlString);
            }
            long result = resultSet.getLong (1);
            resultSet.close();
            if (resultSet.wasNull()) {
                return null;
            } else {
                return dbDomImplementation.retrieveNode(result);
            }
        } catch (SQLException sqle) {
            throw dbDomImplementation.getSqlOrDomException(sqle);
        }
    }

    /** <strong>Not Tested</strong>
     * The number of nodes in this map. The range of valid child node indices 
     * is <code>0</code> to <code>length-1</code> inclusive. 
     */
    public int getLength() {
        try {
            String sqlString = "select getNamedMapLength (?, ?)";
            PreparedStatement statement = (PreparedStatement)dbDomImplementation.getStatements().get(sqlString);
            // if statement is null now, it either isn't present in statements,
            // or is present but uninitialized, which is highly unlikely.
            // putting it into the map wil address both cases
            if (statement == null) {
                statement = con.prepareStatement(sqlString);
                dbDomImplementation.getStatements().put(sqlString, statement);
            }
            statement.setLong (1, elementId);
            statement.setInt  (2, Node.ATTRIBUTE_NODE);
            ResultSet resultSet = statement.executeQuery();
            if (!resultSet.next()) {
                throw new DBDOMSpException (sqlString);
            }
            int result = resultSet.getInt (1);
            resultSet.close();
            return result;
        } catch (SQLException sqle) {
            throw dbDomImplementation.getSqlOrDomException(sqle);
        }
    }

    /**
     * Retrieves a node specified by local name and namespace URI. HTML-only 
     * DOM implementations do not need to implement this method.
     * @param namespaceURIThe namespace URI of the node to retrieve.
     * @param localNameThe local name of the node to retrieve.
     * @return A <code>Node</code> (of any type) with the specified local 
     *   name and namespace URI, or <code>null</code> if they do not 
     *   identify any node in this map.
     * @since DOM Level 2
     *
     * <p>Calls {@link DBElement#getAttributeNodeNS(String, String)}
     */
    public Node getNamedItemNS(String namespaceURI, 
                               String localName) {
        return element.getAttributeNodeNS (namespaceURI, localName);
    }

    /**
     * Adds a node using its <code>namespaceURI</code> and 
     * <code>localName</code>. If a node with that namespace URI and that 
     * local name is already present in this map, it is replaced by the new 
     * one.
     * <br>HTML-only DOM implementations do not need to implement this method.
     * @param argA node to store in this map. The node will later be 
     *   accessible using the value of its <code>namespaceURI</code> and 
     *   <code>localName</code> attributes.
     * @return If the new <code>Node</code> replaces an existing node the 
     *   replaced <code>Node</code> is returned, otherwise <code>null</code> 
     *   is returned.
     * @exception DOMException
     *   WRONG_DOCUMENT_ERR: Raised if <code>arg</code> was created from a 
     *   different document than the one that created this map.
     *   <br>NO_MODIFICATION_ALLOWED_ERR: Raised if this map is readonly.
     *   <br>INUSE_ATTRIBUTE_ERR: Raised if <code>arg</code> is an 
     *   <code>Attr</code> that is already an attribute of another 
     *   <code>Element</code> object. The DOM user must explicitly clone 
     *   <code>Attr</code> nodes to re-use them in other elements.
     * @since DOM Level 2
     *
     * <p>Calls {@link DBElement#setAttributeNodeNS(Attr)}
     */
    public Node setNamedItemNS(Node arg)
                               throws DOMException {
        try {
            return element.setAttributeNodeNS ((Attr)arg);
        } catch (ClassCastException cce) {
            throw new DOMException (DOMException.HIERARCHY_REQUEST_ERR, "Only attributes are allowed in this Map");
        }
    }

    /**
     * Removes a node specified by local name and namespace URI. A removed 
     * attribute may be known to have a default value when this map contains 
     * the attributes attached to an element, as returned by the attributes 
     * attribute of the <code>Node</code> interface. If so, an attribute 
     * immediately appears containing the default value as well as the 
     * corresponding namespace URI, local name, and prefix when applicable.
     * <br>HTML-only DOM implementations do not need to implement this method.
     * @param namespaceURIThe namespace URI of the node to remove.
     * @param localNameThe local name of the node to remove.
     * @return The node removed from this map if a node with such a local 
     *   name and namespace URI exists.
     * @exception DOMException
     *   NOT_FOUND_ERR: Raised if there is no node with the specified 
     *   <code>namespaceURI</code> and <code>localName</code> in this map.
     *   <br>NO_MODIFICATION_ALLOWED_ERR: Raised if this map is readonly.
     * @since DOM Level 2
     */
    public Node removeNamedItemNS(String namespaceURI, 
                                  String localName)
                                  throws DOMException {
        throw new DOMException(DOMException.NOT_SUPPORTED_ERR, "this feature is not implemented yet");
    }

}
