package il.co.iter.dbdom;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;

import org.w3c.dom.NodeList;
import org.w3c.dom.Node;

/**
 * The <code>NodeList</code> interface provides the abstraction of an ordered 
 * collection of nodes, without defining or constraining how this collection 
 * is implemented. <code>NodeList</code> objects in the DOM are live.
 * <p>The items in the <code>NodeList</code> are accessible via an integral 
 * index, starting from 0.
 * <p>See also the <a href='http://www.w3.org/TR/2000/REC-DOM-Level-2-Core-20001113'>Document Object Model (DOM) Level 2 Core Specification</a>.
 *
 * <p>This class represents a <code>Node</code>'s <em>immediate child
 * nodes</em>. For <code>descendants</code>, <code>DBDescendantNodeList</code>
 * is planned.
 */
public class DBChildNodeList implements NodeList {

    private DBNode node;
    private long nodeId;
    private DBDOMImplementation dbDomImplementation;
    private Connection con;

    /**
     * Constructs a {@link DBChildNodeList} associated with a {@link DBNode}
     */
    protected DBChildNodeList (DBNode node) {
        this.node = node;
        this.nodeId = node.getId();
        this.dbDomImplementation = node.getDbDomImplementation();
        this.con = dbDomImplementation.getConnection();
    }

    /**
     * Returns the <code>index</code>th item in the collection. If 
     * <code>index</code> is greater than or equal to the number of nodes in 
     * the list, this returns <code>null</code>.
     * @param indexIndex into the collection.
     * @return The node at the <code>index</code>th position in the 
     *   <code>NodeList</code>, or <code>null</code> if that is not a valid 
     *   index.
     */
    public Node item(int index) {
        try {
            String sqlString = "select childNodeItem (?, ?)";
            PreparedStatement statement = (PreparedStatement)dbDomImplementation.getStatements().get(sqlString);
            // if statement is null now, it either isn't present in statements,
            // or is present but uninitialized, which is highly unlikely.
            // putting it into the map wil address both cases
            if (statement == null) {
                statement = con.prepareStatement(sqlString);
                dbDomImplementation.getStatements().put(sqlString, statement);
            }
            statement.setLong (1, nodeId);
            statement.setInt  (2, index);
            ResultSet resultSet = statement.executeQuery();
            if (!resultSet.next()) {
                throw new DBDOMSpException (sqlString);
            }
            long result = resultSet.getLong (1);
            resultSet.close();
            if (resultSet.wasNull()) {
                return null;
            } else {
                return dbDomImplementation.retrieveNode(result);
            }
        } catch (SQLException sqle) {
            throw dbDomImplementation.getSqlOrDomException(sqle);
        }
    }

    /**
     * The number of nodes in the list. The range of valid child node indices 
     * is 0 to <code>length-1</code> inclusive. 
     */
    public int getLength() {
        try {
            String sqlString = "select getChildNodeListLength (?)";
            PreparedStatement statement = (PreparedStatement)dbDomImplementation.getStatements().get(sqlString);
            // if statement is null now, it either isn't present in statements,
            // or is present but uninitialized, which is highly unlikely.
            // putting it into the map wil address both cases
            if (statement == null) {
                statement = con.prepareStatement(sqlString);
                dbDomImplementation.getStatements().put(sqlString, statement);
            }
            statement.setLong (1, nodeId);
            ResultSet resultSet = statement.executeQuery();
            if (!resultSet.next()) {
                throw new DBDOMSpException (sqlString);
            }
            int result = resultSet.getInt (1);
            resultSet.close();
            return result;
        } catch (SQLException sqle) {
            throw dbDomImplementation.getSqlOrDomException(sqle);
        }
    }

}
