package il.co.iter.dbdom;

import java.sql.Connection;
import java.sql.Statement;
import java.sql.PreparedStatement;
import java.sql.CallableStatement;
import java.sql.ResultSet;
import java.sql.SQLException;

import java.util.Collections;
import java.util.Map;
import java.util.HashMap;
import java.util.WeakHashMap;
import java.util.StringTokenizer;

import org.w3c.dom.DOMImplementation;
import org.w3c.dom.Document;
import org.w3c.dom.DocumentType;
import org.w3c.dom.DOMException;
import org.w3c.dom.Node;
import org.w3c.dom.NamedNodeMap;

 /**
 * The <code>DOMImplementation</code> interface provides a number of methods
 * for performing operations that are independent of any particular instance
 * of the document object model.
 * <p>See also the <a href='http://www.w3.org/TR/2000/REC-DOM-Level-2-Core-20001113'>Document Object Model (DOM) Level 2 Core Specification</a>.
 */
public class DBDOMImplementation implements DOMImplementation {

    //////////////////////////////////////////////////////////////////////////
    // string constants
    //////////////////////////////////////////////////////////////////////////

    //////////////////////////////////////////////////////////////////////////
    // constants that need to be available with or without a DB connection
    //////////////////////////////////////////////////////////////////////////
    private static final String DB_INCOMPATIBLE_MSG =
        "Connection is to a database that is not compatible with DBDOM";

    //////////////////////////////////////////////////////////////////////////
    // constants that can be read from the database
    //////////////////////////////////////////////////////////////////////////
    // used to differentiate between DBDOMExceptions and usual SQLExceptions.
    // see getSqlOrDomException()
    private String domExceptionPrefix;

    //////////////////////////////////////////////////////////////////////////
    // dynamic members
    //////////////////////////////////////////////////////////////////////////
    // reference to the underlying DB
    private Connection connection;
    // PreparedStatements that have been prepared already
    private Map statements;
    // Map of retrieved nodes. It exists to avoid creating a new instance of
    // DBNode every time a method returns a Node. It is a WeakHashMap to allow
    // th egarbage collector to destroy DBNodes that are no longer in use by
    // the application.
    private Map nodes;
    // references to all Document nodes stored in the underlying database
    private DBDocumentNodeMap documents;

    /**
     * Constructs a client connection to a DOM stored in a DBDOM database.
     * <p>The connection is expected to be live, and to a database that has
     * a schema compliant with DBDOM. The constructor tests this.
     */
    public DBDOMImplementation (Connection con) throws DBDOMException {
        try {
            this.connection = con;
            statements = new HashMap();
            // making this Map a normal HashMap may leak memmory, but a Weak
            // map pair is GCed earlyer than needed, resulting in duplicate
            // Nodes. An open issue.
            nodes = new /*Weak*/HashMap();

            if (!dbIsDbdom()) {
                throw new DBDOMException();
                // which will be caught bu the outer try/catch block
            }

            // if the connection is valid, initialize constants with values
            // from the database
            // this uses ad hoc syntax because this is the only place where
            // this table is read; consider creating a stored procedure
            ResultSet rsConstants = connection.createStatement().executeQuery("select value from t_settings where name='domExceptionPrefix'");
            // change this to a while loop if more than one value is read
            rsConstants.next();
            domExceptionPrefix = rsConstants.getString(1);
        } catch (DBDOMException e) {
            throw new DBDOMException (DB_INCOMPATIBLE_MSG);
        } catch (SQLException e) {
            throw new DBDOMException (DB_INCOMPATIBLE_MSG);
        }
    }
	
    /**
     * Test if the DOM implementation implements a specific feature.
     * @param featureThe name of the feature to test (case-insensitive). The
     *   values used by DOM features are defined throughout the DOM Level 2
     *   specifications and listed in the  section. The name must be an XML
     *   name. To avoid possible conflicts, as a convention, names referring
     *   to features defined outside the DOM specification should be made
     *   unique by reversing the name of the Internet domain name of the
     *   person (or the organization that the person belongs to) who defines
     *   the feature, component by component, and using this as a prefix.
     *   For instance, the W3C SVG Working Group defines the feature
     *   "org.w3c.dom.svg".
     * @param versionThis is the version number of the feature to test. In
     *   Level 2, the string can be either "2.0" or "1.0". If the version is
     *   not specified, supporting any version of the feature causes the
     *   method to return <code>true</code>.
     * @return <code>true</code> if the feature is implemented in the
     *   specified version, <code>false</code> otherwise.
     */
    public boolean hasFeature(String feature, String version) {
        try {
            String sqlString = "select hasFeature(?, ?)";
            PreparedStatement statement = (PreparedStatement)statements.get(sqlString);
            // if statement is null now, it either isn't present in statements,
            // or is present but uninitialized, which is highly unlikely.
            // putting it into the map wil address both cases
            if (statement == null) {
                statement = connection.prepareStatement(sqlString);
                statements.put(sqlString, statement);
            }
            statement.setString (1, feature);
            statement.setString (2, version);
            ResultSet resultSet = statement.executeQuery();
            if (!resultSet.next()) {
                throw new DBDOMSpException (sqlString);
            }
            boolean result = resultSet.getBoolean (1);
            resultSet.close();
            return result;
        } catch (SQLException sqle) {
            throw getSqlOrDomException(sqle);
        }
    }

    /**
     * <b>Not Implemented </b>Creates an empty <code>DocumentType</code> node. Entity declarations
     * and notations are not made available. Entity reference expansions and
     * default attribute additions do not occur. It is expected that a
     * future version of the DOM will provide a way for populating a
     * <code>DocumentType</code>.
     * <br>HTML-only DOM implementations do not need to implement this method.
     * @param qualifiedNameThe qualified name of the document type to be
     *   created.
     * @param publicIdThe external subset public identifier.
     * @param systemIdThe external subset system identifier.
     * @return A new <code>DocumentType</code> node with
     *   <code>Node.ownerDocument</code> set to <code>null</code>.
     * @exception DOMException
     *   INVALID_CHARACTER_ERR: Raised if the specified qualified name
     *   contains an illegal character.
     *   <br>NAMESPACE_ERR: Raised if the <code>qualifiedName</code> is
     *   malformed.
     * @since DOM Level 2
     */
    public DocumentType createDocumentType(String qualifiedName,
                                           String publicId,
                                           String systemId)
                                           throws DOMException {
        throw new DOMException(DOMException.NOT_SUPPORTED_ERR, "this feature is not implemented yet");
    }

    /**
     * <b>Not Implemented </b>Creates an XML <code>Document</code> object of the specified type with
     * its document element. HTML-only DOM implementations do not need to
     * implement this method.
     * @param namespaceURIThe namespace URI of the document element to create.
     * @param qualifiedNameThe qualified name of the document element to be
     *   created.
     * @param doctypeThe type of document to be created or <code>null</code>.
     *   When <code>doctype</code> is not <code>null</code>, its
     *   <code>Node.ownerDocument</code> attribute is set to the document
     *   being created.
     * @return A new <code>Document</code> object.
     * @exception DOMException
     *   INVALID_CHARACTER_ERR: Raised if the specified qualified name
     *   contains an illegal character.
     *   <br>NAMESPACE_ERR: Raised if the <code>qualifiedName</code> is
     *   malformed, if the <code>qualifiedName</code> has a prefix and the
     *   <code>namespaceURI</code> is <code>null</code>, or if the
     *   <code>qualifiedName</code> has a prefix that is "xml" and the
     *   <code>namespaceURI</code> is different from "
     *   http://www.w3.org/XML/1998/namespace" .
     *   <br>WRONG_DOCUMENT_ERR: Raised if <code>doctype</code> has already
     *   been used with a different document or was created from a different
     *   implementation.
     * @since DOM Level 2
     */
    public Document createDocument(String namespaceURI,
                                   String qualifiedName,
                                   DocumentType doctype)
                                   throws DOMException {
        throw new DOMException(DOMException.NOT_SUPPORTED_ERR, "this feature is not implemented yet");
    }

    ///////////////////////////////////////////////////////////////////////////
    // private methids
    ///////////////////////////////////////////////////////////////////////////
    /**
     * This method verifies that a connection that was handed to the
     * constructor is to a database with a DBDOM schema. It is very simple now.
     * It should get more comprehensive in the future.
     */
    private boolean dbIsDbdom () {
        return this.hasFeature ("DBDOM", null);
    }

    ///////////////////////////////////////////////////////////////////////////
    // protected methids
    ///////////////////////////////////////////////////////////////////////////
    /**
     * <strong>Non-DOM </strong>Decides if an <code>SQLException</code> was
     * caused by DOMExeption in the database.
     *
     * This method looks at an <code>SQLException</code> and decides if it was
     * thrown by the database because a DOMException occured, or because a
     * genuine SQLException was thrown.
     *
     * <p>Because <code>DOMException</code>s are implemented in the database as
     * usual exceptions, the only way to let the client know that a
     * DOMException occured is to include a flag in the message of the
     * exception.
     *
     * <p>DBDOMExceptions include a delimiter in their message
     * the format of the message is
     * &lt;domExceptionPrefix&gt; &lt;code&gt; &lt;message&gt;
     *
     * <p>The value of <code>domExceptionPrefix</code> is read from the
     * database when this object is constructed</p>
     *
     * <p>Note: although it would make sense for this method to be in {@link
     * DBDOMException}, it is inadvisable, because
     * <code>domExceptionPrefix</code> needs to be read from the database.
     * {@link DBDOMException} does not hold a refernce to the database (let
     * alone a static one), nor to {@link DBDOMImplementation}.</p>
     *
     * @param sqle <code>SQLException</code> to be examined.
     * @return either a {@link org.w3c.dom.DOMException} or an SQLException.
     */
    protected RuntimeException getSqlOrDomException (SQLException sqle) {
        StringTokenizer exceptionMsg = new StringTokenizer (sqle.getMessage());
        while (exceptionMsg.hasMoreElements()) {
            // if the connection is not to a DBDOM database, domExceptionPrefix
            // will be null, and so will never match anything.
            if (exceptionMsg.nextElement().equals(domExceptionPrefix)) {
                return new DOMException (Short.parseShort(exceptionMsg.nextElement().toString()),
                                         exceptionMsg.nextElement().toString());
            }
        }
        // if the loot exited, domExceptionPrefix wasn't found
        // the exception is a regular SQLException
        return new DBDOMSqlException(sqle);
    }

    /**
     * <strong>Non-DOM </strong>Returns a reference to the database connection.
     */
    protected Connection getConnection() {
        return this.connection;
    }

    /**
     * <strong>Non-DOM </strong>Returns a map of all known statements.
     */
    protected Map getStatements() {
        return statements;
    }

    /**
     * <strong>Non-DOM </strong>Attmpts to read a node from the database.
     *
     * <p>References to retrieved {@link DBNode}s are kept in a private
     * {@link WeakHashMap}. First, this method tries to get the {@link DBNode}
     * from the {@link WeakHashMap}; if that fails, a JDBC read is performed.
     *
     * <p>This method's retrun type may need to be modified to {@link DBNode}.
     *
     * @param id id of the Node in the database
     * @return a {@link DBNode}
     */
    protected Node retrieveNode(long id) {
        Long longId = new Long(id);
        DBNode node = (DBNode)nodes.get(longId);
        // consider checking nodeExsits() in any case, not just when it is not
        // found in WeakHasMap, and removing it from Map if it doesn't.
        // This will help detect cases where a Node has been deleted from the
        // database, but has not yet been garabage collected, either because GC
        // hasn't got around to it yet, or because the application holds a
        // reference to it.
        if (node == null) {
            short nodeType = 0;  // invalid node type
            try {
                String sqlString = "select getNodeType (?)";
                PreparedStatement statement = (PreparedStatement)statements.get(sqlString);
                // if statement is null now, it either isn't present in statements,
                // or is present but uninitialized, which is highly unlikely.
                // putting it into the map wil address both cases
                if (statement == null) {
                    statement = connection.prepareStatement(sqlString);
                    statements.put(sqlString, statement);
                }
                statement.setLong (1, id);
                ResultSet resultSet = statement.executeQuery();
                if (!resultSet.next()) {
                    throw new DBDOMSpException (sqlString);
                }
                nodeType = resultSet.getShort (1);
                resultSet.close();

                // this is different from most DB access routines
                switch (nodeType) {
                    case Node.ELEMENT_NODE:
                        node = new DBElement(this, id);
                        break;
                    case Node.ATTRIBUTE_NODE:
                        node = new DBAttr(this, id);
                        break;
                    case Node.TEXT_NODE:
                        node = new DBText(this, id);
                        break;
                    case Node.CDATA_SECTION_NODE:
                        node = new DBCDATASection(this, id);
                        break;
                    case Node.ENTITY_REFERENCE_NODE:
                        // replace once the constructor is available
                        node = new DBNode(this, id);
                        break;
                    case Node.ENTITY_NODE:
                        // replace once the constructor is available
                        node = new DBNode(this, id);
                        break;
                    case Node.PROCESSING_INSTRUCTION_NODE:
                        node = new DBProcessingInstruction(this, id);
                        break;
                    case Node.COMMENT_NODE:
                        node = new DBComment(this, id);
                        break;
                    case Node.DOCUMENT_NODE:
                        node = new DBDocument(this, id);
                        break;
                    case Node.DOCUMENT_TYPE_NODE:
                        node = new DBDocumentType(this, id);
                        break;
                    case Node.DOCUMENT_FRAGMENT_NODE:
                        // replace once the constructor is available
                        node = new DBNode(this, id);
                        break;
                    case Node.NOTATION_NODE:
                        // replace once the constructor is available
                        node = new DBNode(this, id);
                        break;
                    default:
                        throw new DBDOMException ("Internal error: unknown NodeType: " + nodeType);
                }
                nodes.put(longId, node);
            } catch (SQLException sqle) {
                throw getSqlOrDomException(sqle);
            }
        }
        return node;
    }

    /**
     * <strong>Non-DOM </strong>Tests whether or not a node with the specified
     * id is found in the database
     */
    protected boolean nodeExists(long id) {
        try {
            String sqlString = "select nodeExists (?)";
            PreparedStatement statement = (PreparedStatement)statements.get(sqlString);
            // if statement is null now, it either isn't present in statements,
            // or is present but uninitialized, which is highly unlikely.
            // putting it into the map wil address both cases
            if (statement == null) {
                statement = connection.prepareStatement(sqlString);
                statements.put(sqlString, statement);
            }
            statement.setLong (1, id);
            ResultSet resultSet = statement.executeQuery();
            if (!resultSet.next()) {
                throw new DBDOMSpException (sqlString);
            }
            boolean result = resultSet.getBoolean (1);
            resultSet.close();
            return result;
        } catch (SQLException sqle) {
            throw getSqlOrDomException(sqle);
        }
    }

    //////////////////////////////////////////////////////////////////////////
    // public non-DOM methods
    //////////////////////////////////////////////////////////////////////////

    /**
     * <strong>Non-DOM</strong> Cretes a new document in the database. This
     * document can be passed as a parameter to
     * {@link il.co.iter.dbdom.parsers.DBDOMParser} to be populated.
     */
    public DBDocument createNamedDocument (String name) {
        try {
            String sqlString = "select createNamedDocument (?)";
            PreparedStatement statement = (PreparedStatement)statements.get(sqlString);
            // if statement is null now, it either isn't present in statements,
            // or is present but uninitialized, which is highly unlikely.
            // putting it into the map wil address both cases
            if (statement == null) {
                statement = connection.prepareStatement(sqlString);
                statements.put(sqlString, statement);
            }
            statement.setString (1, name);
            ResultSet resultSet = statement.executeQuery();
            if (!resultSet.next()) {
                throw new DBDOMSpException (sqlString);
            }
            long result = resultSet.getLong (1);
            resultSet.close();
            if (resultSet.wasNull()) {
                return null;
            } else {
                return (DBDocument)retrieveNode(result);
            }
        } catch (SQLException sqle) {
            throw getSqlOrDomException(sqle);
        }
    }
    /**
     * <string>Non-DOM<strong> Returns a map of all documents stored in a
     * DBDOM database.
     */
    public NamedNodeMap getDocuments() {
        if (documents == null) {
            documents = new DBDocumentNodeMap (this);
        }
        return documents;
    }

    /**
     * <strong>Non-DOM</strong> Encapsulates {@link java.sql.Connection#setAutoCommit(boolean)}
     */
    public void setAutoCommit(boolean autoCommit) {
        try {
            connection.setAutoCommit (autoCommit);
        } catch (SQLException sqle) {
            throw getSqlOrDomException(sqle);
        }
    }

    /**
     * <strong>Non-DOM</strong> Encapsulates {@link java.sql.Connection#getAutoCommit()}
     */
    public boolean getAutoCommit() {
        try {
            return connection.getAutoCommit ();
        } catch (SQLException sqle) {
            throw getSqlOrDomException(sqle);
        }
    }

    /**
     * <strong>Non-DOM</strong> Encapsulates {@link java.sql.Connection#commit()}
     */
    public void commit() {
        try {
            connection.commit ();
        } catch (SQLException sqle) {
            throw getSqlOrDomException(sqle);
        }
    }

    /**
     * <strong>Non-DOM</strong> Encapsulates {@link java.sql.Connection#rollback()}
     */
    public void rollback() {
        try {
            connection.rollback ();
        } catch (SQLException sqle) {
            throw getSqlOrDomException(sqle);
        }
    }
}

