package il.co.iter.dbdom;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;

import org.w3c.dom.ProcessingInstruction;
import org.w3c.dom.Node;
import org.w3c.dom.DOMException;

/**
 * The <code>ProcessingInstruction</code> interface represents a "processing 
 * instruction", used in XML as a way to keep processor-specific information 
 * in the text of the document.
 * <p>See also the <a href='http://www.w3.org/TR/2000/REC-DOM-Level-2-Core-20001113'>Document Object Model (DOM) Level 2 Core Specification</a>.
 */
public class DBProcessingInstruction extends DBNode implements ProcessingInstruction {

    /**
     * Calls super(dbDomImplementation, id)
     */
    protected DBProcessingInstruction (DBDOMImplementation dbDomImplementation, long id) {
        super(dbDomImplementation, id);
    }
    /**
     * The target of this processing instruction. XML defines this as being 
     * the first token following the markup that begins the processing 
     * instruction.
     */
    public String getTarget() {
        try {
            String sqlString = "select getTarget (?)";
            PreparedStatement statement = (PreparedStatement)dbDomImplementation.getStatements().get(sqlString);
            // if statement is null now, it either isn't present in statements,
            // or is present but uninitialized, which is highly unlikely.
            // putting it into the map wil address both cases
            if (statement == null) {
                statement = con.prepareStatement(sqlString);
                dbDomImplementation.getStatements().put(sqlString, statement);
            }
            statement.setLong (1, super.getId());
            ResultSet resultSet = statement.executeQuery();
            if (!resultSet.next()) {
                throw new DBDOMSpException (sqlString);
            }
            String result = resultSet.getString (1);
            resultSet.close();
            return result;
        } catch (SQLException sqle) {
            throw dbDomImplementation.getSqlOrDomException(sqle);
        }
    }

    /**
     * The content of this processing instruction. This is from the first non 
     * white space character after the target to the character immediately 
     * preceding the <code>?&gt;</code>.
     * @exception DOMException
     *   NO_MODIFICATION_ALLOWED_ERR: Raised when the node is readonly.
     *
     * <p>This attribute is named PiData in the database to avoid name clash
     * with interface {@link org.w3c.dom.CharacterData}
     */
    public String getData() {
        try {
            String sqlString = "select getPiData(?)";
            PreparedStatement statement = (PreparedStatement)dbDomImplementation.getStatements().get(sqlString);
            // if statement is null now, it either isn't present in statements,
            // or is present but uninitialized, which is highly unlikely.
            // putting it into the map wil address both cases
            if (statement == null) {
                statement = con.prepareStatement(sqlString);
                dbDomImplementation.getStatements().put(sqlString, statement);
            }
            statement.setLong (1, super.getId());
            ResultSet resultSet = statement.executeQuery();
            if (!resultSet.next()) {
                throw new DBDOMSpException (sqlString);
            }
            String result = resultSet.getString (1);
            resultSet.close();
            return result;
        } catch (SQLException sqle) {
            throw dbDomImplementation.getSqlOrDomException(sqle);
        }
    }

    public void setData(String data)
                          throws DOMException {
        try {
            String sqlString = "select setPiData (?, ?)";
            PreparedStatement statement = (PreparedStatement)dbDomImplementation.getStatements().get(sqlString);
            // if statement is null now, it either isn't present in statements,
            // or is present but uninitialized, which is highly unlikely.
            // putting it into the map wil address both cases
            if (statement == null) {
                statement = con.prepareStatement(sqlString);
                dbDomImplementation.getStatements().put(sqlString, statement);
            }
            statement.setLong   (1, super.getId());
            statement.setString (2, data);
            ResultSet resultSet = statement.executeQuery();
            if (!resultSet.next()) {
                throw new DBDOMSpException (sqlString);
            }
            resultSet.close();
        } catch (SQLException sqle) {
            throw dbDomImplementation.getSqlOrDomException(sqle);
        }
    }

}
