package il.co.iter.dbdom;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;

/**
 * Represents a stored procedure with a known result set signature in the
 * underlying databas.&nbsp;<b>This class is not currently used.</b>  Some of
 * the comments in the class are still highly relevant.
 *
 * <p>DBDOM makes extensive use of stored procedures. This class exists for
 * two reasons:
 * <ol>
 *   <li>To hide the difference betbeen {@link PreparedStatement} and
 *       {@link java.sql.CallableStatement}. The choice depends on whether or
 *       not support is available in a particular JDBC driver.
 *   <li>To encapsulate common operations:
 *     <ul>
 *       <li>wrapping {@link SQLException}s in {@link DBDOMSqlException}s,
 *          because DOM methods can't raise exceptions other than
 *          {@link org.w3c.dom.DOMException};
 *       <li>initializing (preparing) {@link java.sql.Statement}s on demand;
 *       <li>checking that SPs return expected {@link ResultSet}s.
 *     </ul>
 * </ol>
 *
 * {@link java.sql.Statement}s (or their subclasses) that represent the stored
 * procedures are kept as private members of {@link DBDOMImplementation}.
 * They are not initialized in {@link DBDOMImplementation}'s constructor, but
 * rather on first access.
 *
 * <p>Keeping {@link java.sql.Statement}s in {@link DBDOMImplementation}
 * optimizes for size, since individual
 * {@link DBNode}s don't need to cary their own statements. This may become
 * a performance bottleneck: only one <code>Node</code> at a time is able to
 * execute a particular DOM method. If this in fact bocomes an issue, move
 * {@link StoredProcedure} declarations to {@link DBNode} and it's subclasses.
 *
 * <p>{@link java.sql.Statement}s cannot be static members, because different
 * instances of {@link DBDOMImplementation} may hold connections to different
 * databases - with different {@link java.sql.Connection} objects and possibly
 * different versions of DBDOM and possibly different RDBMSs altogether.
 *
 * <p>All DBDOM stored procedures
 * have a known type of result set. If a stored procedure returns a different
 * type of result set than expected, it means that the database is
 * incompatible with DBDOM or that there is a version missmatch between the
 * database and the Java wrappers. See {@link DBDOMSpException}
 *
 * <p>{@link StoredProcedure} may be subclassed to emulate stored procedures
 * in an RDMS that doesn't provide a native PS mechanism.
 */
public class StoredProcedure {
    /**
     * Specifies that a stored procedure may return an unlimited number of
     * results per invocation.
     */
    public static final int UNLIMITED = -1;

    private Connection connection;
    private PreparedStatement statement;
    private ResultSet resultSet;
    // name of the SQL stored procedure
    private String spName;
    //SQL string that would be used to run this SP against the databse
    private String sqlStr;

    protected StoredProcedure (DBDOMImplementation dbDomImplementation,
                               String spName,
                               int nOfParams,
                               int nOfResults){

        connection = dbDomImplementation.getConnection();
        this.spName = spName;

        // construct sql string
        // consider movig this block into a separate method for clarity,
        // and to allow overriding by subclasses
        sqlStr = "select " + spName + '(';
        // the loop runs nOfParams-1 times, to account for the last comma,
        // and the case of no paramaters
        // if nOfParams is negative, the loop and the following condition
        // aren't entered, and so it is treated as 0
        for (int i=1; i < nOfParams; i++) {
            sqlStr += "?, ";
        }
        if (nOfParams > 0) {
            sqlStr += '?';
        }
        sqlStr += ')';
    }

    /**
     * This simple operation is implemented as a protected method to allow
     * overriding by subclases.
     *
     * <p>In particular, a subclass that uses a
     * {@link java.sql.CallableStatement} as opposed to
     * {@link PreparedStatement} should override this method.
     */
    protected void initStatementIfNull () throws SQLException {
        if (statement == null) {
            statement = connection.prepareStatement (sqlStr);
        }
    }

    protected ResultSet executeQuery() {
        // null ststement arguments before exiting
        try {
            initStatementIfNull();
            return statement.executeQuery();
        } catch (SQLException sqle) {
            throw new DBDOMSqlException (sqle);
        }
    }

    //////////////////////////////////////////////////////////////////////////
    // statement.setSomeType methods
    //////////////////////////////////////////////////////////////////////////

    protected void setInt (int parameterIndex, int x) {
        try {
            initStatementIfNull();
            statement.setInt (parameterIndex, x);
        } catch (SQLException sqle) {
            throw new DBDOMSqlException (sqle);
        }
    }

    protected void setString (int parameterIndex, String x) {
        try {
            initStatementIfNull();
            statement.setString (parameterIndex, x);
        } catch (SQLException sqle) {
            throw new DBDOMSqlException (sqle);
        }
    }

    protected void setNull(int parameterIndex, int sqlType) {
        try {
            initStatementIfNull();
            statement.setNull (parameterIndex, sqlType);
        } catch (SQLException sqle) {
            throw new DBDOMSqlException (sqle);
        }
    }

    protected void setNull(int paramIndex, int sqlType, String typeName) {
        try {
            initStatementIfNull();
            statement.setNull (paramIndex, sqlType, typeName);
        } catch (SQLException sqle) {
            throw new DBDOMSqlException (sqle);
        }
    }

    /**
     * Returns the SQL string that would be used to run this SP against the
     * databse.
     */
    public String toString() {
        return sqlStr;
    }

    /**
     * Returns the name of the SQL stored procedure
     */
    public String getSpName() {
        return spName;
    }

    /**
     * Two {@link StoredProcedure}s are equal if and only if their stored
     * procedure names are equal.
     *
     * <p>Consider cheching that Connections are equal, too.
     */
    /*
    public boolean equals(Object obj) {
    	if (this == obj) {
    	    return true;
    	}
    	
        if (obj.getClass() == this.getClass()) {
            if (this.getSpName().equals(((StoredProcedure)obj).getSpName())) {
                return true;
            }
    	}
    	
    	return false;
    } */ // equals(Object)
}  // StoredProcedure
