--------------------------------------------------------------------------
-- Private functions
--------------------------------------------------------------------------
CREATE OR REPLACE FUNCTION	nodeExists	(nodeId NUMBER)	RETURN BOOLEAN AS
	nOfNodesFound	NUMBER := 0;
BEGIN
	SELECT	COUNT (id)
	INTO	nOfNodesFound
	FROM	t_Node
	WHERE	id = nodeId;
	
	IF nOfNodesFound = 0 THEN
		-- this ad-hoc error should later go into t_DBDOMException
		raise_application_error (-20101, 'node ' || nodeId || ' does not exist');
		RETURN FALSE;
	ELSIF nOfNodesFound > 1 THEN
		-- should never happen because of primary key constraints; consider removing to improve performance
		raise_application_error (-20102, 'corrupted data in the database: node ' || nodeId || ' is found ' || nOfNOdesFound || ' times; expected once only');
		RETURN FALSE;
	ELSE
		RETURN TRUE;
	END IF;
END nodeExists;

/


CREATE OR REPLACE FUNCTION	nodeTypeMatches	(nodeId NUMBER, expectedNodeType NUMBER) RETURN BOOLEAN AS
-- consider merging thses two functions into one: it is not clear that there will be any use for a standalone nodeExists() function; also, performance should improve, because only a single lookup will be needed
	actualNodeType	NUMBER;
BEGIN
	IF NOT nodeExists (nodeId) THEN
		RETURN FALSE;
	ELSE
		SELECT	nodeType
		INTO	actualNodeType
		FROM	t_Node
		WHERE	id = nodeId;
		
		IF expectedNodeType != actualNodeType THEN
			raise_application_error (-20103, 'type mismatch: node is of type ' || actualNodeType || ', expected ' || expectedNodeType);
			RETURN FALSE;
		ELSE
			RETURN TRUE;
		END IF;
	END IF;
END nodeTYpeMatches;

/


-- a stub only; always returns TRUE
CREATE OR REPLACE FUNCTION isValidName (theString VARCHAR) RETURN BOOLEAN AS
BEGIN
	RETURN TRUE;
END isValidName;

/


--------------------------------------------------------------------------
-- Exception DOMException
--------------------------------------------------------------------------
CREATE OR REPLACE PROCEDURE	DOMException	(code NUMBER) AS
	message		VARCHAR (255);
BEGIN
	SELECT	msg
	INTO	message
	FROM	t_DOMException
	WHERE	id = code;
	
	IF message IS NULL THEN
		raise_application_error (-20000, 'error raising DOMException, invalid exception code' || code);
	ELSE
		raise_application_error (-20000 - code, message);
	END IF;
END DOMException;

/


--------------------------------------------------------------------------
-- Interface DOMImplementation
--------------------------------------------------------------------------
CREATE OR REPLACE FUNCTION	hasFeature (arg_feature VARCHAR, arg_version VARCHAR)	RETURN BOOLEAN AS
	nOfConformingRecords	NUMBER := 0;
BEGIN

	IF arg_version IS NOT NULL THEN
		SELECT	COUNT (*)
		INTO	nOfConformingRecords
		FROM	t_DOMImplementation
		WHERE	UPPER (feature) = UPPER (arg_feature) AND
			UPPER (version) = UPPER (arg_version);
	ELSE
		SELECT	COUNT (*)
		INTO	nOfConformingRecords
		FROM	t_DOMImplementation
		WHERE	UPPER (feature) = UPPER (arg_feature);
	END IF;
	
	RETURN	nOfConformingRecords > 0;
END hasFeature;

/


-- creates an empty, DTD-less document
-- unspecified in DOM level 1; propriatory extention
CREATE OR REPLACE FUNCTION	createNewDocument RETURN NUMBER AS
	PRAGMA		AUTONOMOUS_TRANSACTION;	-- ?
	newDocumentId	NUMBER;
BEGIN
	SELECT	sq_Node.nextval
	INTO	newDocumentId
	FROM	DUAL;
	
	INSERT	INTO
	t_Node	(id, nodeType, ownerDocument)
	VALUES	(newDocumentId, 9, NULL);
	
	COMMIT;
	
	RETURN newDocumentId;
END createNewDocument;

/


--------------------------------------------------------------------------
-- Interface DocumentFragment
--------------------------------------------------------------------------


--------------------------------------------------------------------------
-- Interface Document
--------------------------------------------------------------------------

-- readonly attribute  DocumentType         doctype;
CREATE OR REPLACE FUNCTION	getDocumentType (nodeId NUMBER) RETURN NUMBER AS
	documentTypeNodeId	NUMBER;
BEGIN
	--if type checking is ommited, and nodeId reffers to a wrong type of Node, NULL will be returned.
	IF nodeTypeMatches (nodeId, 9) THEN
		SELECT	child_node_id
		INTO	documentTypeNodeId
		FROM	t_arch,
			t_archType
		WHERE	t_arch.type_id = t_archType.id AND
			t_arch.parent_node_id = nodeId AND
			t_archtype.child_node_type_id = 10;
	END IF;
	
	RETURN documentTypeNodeId;
END getDocumentType;

/


-- readonly attribute  DOMImplementation    implementation;
-- "The DOMImplementation interface provides a number of methods for performing operations that are independent of any particular instance of the document object model."
-- The only method DOMImplementation has in Level 1 is hasFeature. There is no object to reference in the database.


-- readonly attribute  Element              documentElement;
-- this function hasn't been tested yet for the lack of data
CREATE OR REPLACE FUNCTION getDocumentElement (nodeId NUMBER) RETURN NUMBER AS
	documentElementId	NUMBER;
BEGIN
	SELECT	child_node_id
	INTO	documentElementId
	FROM	t_arch,
		t_archType
	WHERE	t_arch.type_id = t_archType.id AND
		t_arch.parent_node_id = nodeId AND
		t_archType.child_node_type_id = 1;
	
END getDocumentElement;

/


-- private funcion to create the t_Node part of a new node
CREATE OR REPLACE FUNCTION createNode (parentDocumentId NUMBER, nodeTypeId NUMBER) RETURN NUMBER AS
	newNodeId	NUMBER;
BEGIN
	IF nodeTypeMatches (parentDocumentId, 9) THEN
		INSERT	INTO
		t_Node	(id, nodeType, ownerDocument)
		VALUES	(sq_Node.nextval, nodeTypeId, parentDocumentId);
		-- if nodeTypeId is invalid, FK will be violated and an exception raised, although not a very descriptive one
		
		SELECT	sq_Node.currval
		INTO	newNodeID
		FROM	DUAL;
	END IF;
	RETURN newNodeId;
END createNode;

/


-- Element                   createElement(in DOMString tagName)
-- raises(DOMException);
CREATE OR REPLACE FUNCTION createElement (parentDocumentId NUMBER, tagName VARCHAR) RETURN NUMBER AS
	PRAGMA		AUTONOMOUS_TRANSACTION;		-- ?
	newElementId	NUMBER;
	tagId		NUMBER;
	nOfConformingTags	NUMBER;
BEGIN
	IF nodeTypeMatches (parentDocumentId, 9) THEN
		-- check that the tagName does not contain illegal characters
		IF isValidName (tagName) THEN
			
			-- get tagId
			-- rewrite using an explicit cursor
			SELECT	COUNT (id)
			INTO	nOfConformingTags
			FROM	t_tag
			WHERE	name = tagName;
			
			IF nOfConformingTags > 0 THEN
				SELECT	id
				INTO	tagId
				FROM	t_tag
				WHERE	name = tagName;
			ELSE
				INSERT	INTO
				t_tag	(id, name)
				VALUES	(sq_Tag.nextval, tagName);
				
				SELECT	sq_Tag.currval
				INTO	tagId
				FROM	DUAL;
			END IF;
			
			-- create the Node part
			newElementId := createNode (parentDocumentId, 1);
		
			-- create the Element part
			INSERT	INTO
			t_Element (id, tag_id)
			VALUES	(newElementId, tagId);
			
			COMMIT;
		END IF;
	END IF;
	
	RETURN newElementId;
END createElement;

/


-- Creates an empty DocumentFragment object
CREATE OR REPLACE FUNCTION createDocumentFragment (parentDocumentId NUMBER) RETURN NUMBER AS
	PRAGMA		AUTONOMOUS_TRANSACTION;		-- ?
	newDocumentFragmentId	NUMBER;
BEGIN
	IF nodeTypeMatches (parentDocumentId, 9) THEN
		-- create the Node part
		newDocumentFragmentId := createNode (parentDocumentId, 11);
		
		-- create the DocumentFragmnet part
		INSERT	INTO
		t_DocumentFragment	(id)
		VALUES	(newDocumentFragmentId);
		
		COMMIT;
	END IF;
	
	RETURN newDocumentFragmentId;
END createDocumentFragment;

/


-- Creates a Text node given the specified string
CREATE OR REPLACE FUNCTION createTextNode (parentDocumentId NUMBER, argData VARCHAR) RETURN INT AS
	PRAGMA		AUTONOMOUS_TRANSACTION;		-- ?
	newTextNodeId	NUMBER;
BEGIN
	IF nodeTypeMatches (parentDocumentId, 9) THEN
		-- create the Node part
		newTextNodeId := createNode (parentDocumentId, 3);
		
		-- create the Text part
		INSERT	INTO
		t_CharacterData	(id, data)
		VALUES	(newTextNodeId, argData);
		
		COMMIT;
	END IF;
	
	RETURN	newTextNodeId;
END createTextNode;

/


-- Creates a Comment node given the specified string
CREATE OR REPLACE FUNCTION createComment (parentDocumentId NUMBER, argData VARCHAR) RETURN INT AS
	PRAGMA		AUTONOMOUS_TRANSACTION;		-- ?
	newCommentId	NUMBER;
BEGIN
	IF nodeTypeMatches (parentDocumentId, 9) THEN
		-- create the Node part
		newCommentId := createNode (parentDocumentId, 8);
		
		-- create the Text part
		INSERT	INTO
		t_CharacterData	(id, data)
		VALUES	(newCommentId, argData);
		
		COMMIT;
	END IF;
	
	RETURN	newCommentId;
END createComment;

/


-- Creates a CDATASection node given the specified string
CREATE OR REPLACE FUNCTION createCDATASection (parentDocumentId NUMBER, argData VARCHAR) RETURN INT AS
	PRAGMA			AUTONOMOUS_TRANSACTION;		-- ?
	newCDATASectionId	NUMBER;
BEGIN
	IF nodeTypeMatches (parentDocumentId, 9) THEN
		-- create the Node part
		newCDATASectionId := createNode (parentDocumentId, 4);
		
		-- create the Text part
		INSERT	INTO
		t_CharacterData	(id, data)
		VALUES	(newCDATASectionId, argData);
		
		COMMIT;
	END IF;
	
	RETURN	newCDATASectionId;
END createCDATASection;

/


-- ProcessingInstruction     createProcessingInstruction(in DOMString target,
-- in DOMString data)
-- raises(DOMException);
CREATE OR REPLACE FUNCTION createProcessingInstruction (parentDocumentId NUMBER, argTarget VARCHAR, argData VARCHAR) RETURN NUMBER AS
	PRAGMA				AUTONOMOUS_TRANSACTION;
	newProcessingInstructionId	NUMBER;
BEGIN
	IF nodeTypeMatches (parentDocumentId, 9) THEN
		--create the Node part
		newProcessingInstructionId := createNode (parentDocumentId, 7);
		
		-- create the ProcessingInstruction part
		INSERT	INTO
		t_ProcessingInstruction (id, target, data)
		VALUES	(newProcessingInstructionId, argTarget, argData);
		
		COMMIT;
	END IF;
			
	RETURN newProcessingInstructionId;
END createProcessingInstruction;

/


-- Attr                      createAttribute(in DOMString name)
-- raises(DOMException);
CREATE OR REPLACE FUNCTION createAttribute (parentDocumentId NUMBER, attrName VARCHAR) RETURN NUMBER AS
	PRAGMA			AUTONOMOUS_TRANSACTION;
	newAttributeId		NUMBER;
	attrNameId		NUMBER;
	nOfConformingAttribs	NUMBER;
BEGIN
	IF nodeTypeMatches (parentDocumentId, 9) THEN
		-- check that the attrName does not contain illegal characters
		IF isValidName (attrName) THEN

			-- get attrNameId
			-- rewrite using an explicit cursor
			SELECT	COUNT (id)
			INTO	nOfConformingAttribs
			FROM	t_attrName
			WHERE	name = attrName;
			
			IF nOfConformingattribs > 0 THEN
				SELECT	id
				INTO	attrNameId
				FROM	t_attrName
				WHERE	name = attrName;
			ELSE
				INSERT		INTO
				t_attrName	(id, name)
				VALUES		(sq_attrName.nextval, attrName);
				
				SELECT	sq_attrName.currval
				INTO	attrNameId
				FROM	DUAL;
			END IF;
			
			-- create the Node part
			newAttributeId := createNode (parentDocumentId, 2);
		
			-- create the Attribute part
			INSERT	INTO
			t_Attr	(id, name_id, specified)
			VALUES	(newAttributeId, attrNameId, 1);
			
			COMMIT;
		END IF;
	END IF;
	
	RETURN newAttributeId;
END createAttribute;

/


-- EntityReference           createEntityReference(in DOMString name)
-- raises(DOMException);
CREATE OR REPLACE FUNCTION createEntityReference (parentDocumentId NUMBER, name VARCHAR) RETURN NUMBER AS
BEGIN
	raise_application_error (-20104, 'createEntityReference() not implemented yet');
	RETURN NULL;
END createEntityReference;

/


-- NodeList                  getElementsByTagName(in DOMString tagname);

-- appearently, this it the same method as Element.getElementsByTagName
-- see comment in Element.getElementsByTagName


--------------------------------------------------------------------------
-- Interface Node
--------------------------------------------------------------------------
-- readonly attribute  DOMString            nodeName;
CREATE OR REPLACE FUNCTION getNodeName (nodeId NUMBER) RETURN VARCHAR AS
	nodeTypeId	NUMBER;
	result		VARCHAR (255);
BEGIN
	SELECT	NodeType
	INTO	nodeTypeId
	FROM	t_node
	WHERE	id = nodeId;
	-- rewrite this select as an explicit cursor. this will enable catching nonexisting nodeIds
	-- the current implementation will silently exit (returning NULL) if no matching Nodes is found

	IF	nodeTypeId = 1	-- Element
	THEN	SELECT	t_tag.name
		INTO	result
		FROM	t_Element,
			t_tag
		WHERE	t_Element.tag_id = t_tag.id AND
			t_Element.id     = nodeId;
			
	ELSIF	nodeTypeId = 2	-- Attr
	THEN	SELECT	t_attrName.name
		INTO	result
		FROM	t_Attr,
			t_attrName
		WHERE	t_Attr.name_id = t_attrName.id AND
			t_Attr.id = nodeId;
		
	ELSIF	nodeTypeId = 3	-- Text
	THEN	result := '#text';
	
	ELSIF	nodeTypeId = 4	-- CDATA
	THEN	result := '#cdata-section';
	
	ELSIF	nodeTypeId = 5	-- ENTITY_REFERENCE
	THEN	raise_application_error (-20000, 'entity references are not handled yet');
	
	ELSIF	nodeTypeId = 6	-- ENTITY
	THEN	SELECT	name
		INTO	result
		FROM	t_Entity
		WHERE	id = nodeId;
	
	ELSIF	nodeTypeId = 7	-- PROCESSING_INSTRUCTION
	THEN	SELECT	target
		INTO	result
		FROM	t_ProcessingInstruction
		WHERE	id = nodeId;
	
	ELSIF	nodeTypeId = 8	-- Comment
	THEN	result := '#comment';
	
	ELSIF	nodeTypeId = 9	-- Document
	THEN	result := '#document';
	
	ELSIF	nodeTypeId = 10	-- DocumentType
	THEN	SELECT	name
		INTO	result
		FROM	t_DocumentType
		WHERE	id = nodeId;
	
	ELSIF	nodeTypeId = 11	-- DocumentFragment
	THEN	result := '#document-fragment';
		
	ELSIF	nodeTypeId = 12	-- NOTATION
	THEN	SELECT	name
		INTO	result
		FROM	t_Notation
		WHERE	id = nodeId;
	
	ELSE
		raise_application_error (-20000, 'internal error in getNodeName, corrupted data in the database: nodeId ' || nodeId || ': unkmown nodeType');
	END IF;
	
	RETURN result;		

END getNodeName;

/


-- attribute  DOMString            nodeValue;
-- // raises(DOMException) on setting
-- // raises(DOMException) on retrieval
CREATE OR REPLACE FUNCTION getNodeValue (nodeId NUMBER) RETURN VARCHAR AS
	nodeTypeId	NUMBER;
	result		VARCHAR (255);
	CURSOR		cr_NodeType	IS SELECT NodeType FROM t_Node WHERE id = nodeId;
BEGIN

	OPEN	cr_NodeType;
	FETCH	cr_NodeType INTO nodeTypeId;
	IF cr_NodeType%NOTFOUND THEN
		raise_application_error (-20000, 'Node ' || nodeId || ' does not exist in the database');
	END IF;
	CLOSE	cr_NodeType;

	IF	nodeTypeId = 1	-- Element
	THEN	result := NULL;
			
	ELSIF	nodeTypeId = 2	-- Attr
	THEN	SELECT	value
		INTO	result
		FROM	t_Attr
		WHERE	id = nodeId;
		
	ELSIF	nodeTypeId = 3	-- Text
	THEN	SELECT	data
		INTO	result
		FROM	t_CharacterData
		WHERE	id = nodeId;
	
	ELSIF	nodeTypeId = 4	-- CDATA
	THEN	SELECT	data
		INTO	result
		FROM	t_CharacterData
		WHERE	id = nodeId;
	
	ELSIF	nodeTypeId = 5	-- ENTITY_REFERENCE
	THEN	result := NULL;
	
	ELSIF	nodeTypeId = 6	-- ENTITY
	THEN	result := NULL;
	
	ELSIF	nodeTypeId = 7	-- PROCESSING_INSTRUCTION
	THEN	SELECT	Data
		INTO	result
		FROM	t_ProcessingInstruction
		WHERE	id = nodeId;
	
	ELSIF	nodeTypeId = 8	-- Comment
	THEN	SELECT	data
		INTO	result
		FROM	t_CharacterData
		WHERE	id = nodeId;
	
	ELSIF	nodeTypeId = 9	-- Document
	THEN	result := NULL;
	
	ELSIF	nodeTypeId = 10	-- DocumentType
	THEN	result := NULL;
	
	ELSIF	nodeTypeId = 11	-- DocumentFragment
	THEN	result := NULL;
		
	ELSIF	nodeTypeId = 12	-- NOTATION
	THEN	result := NULL;
	
	ELSE
		raise_application_error (-20000, 'internal error in getNodeName, corrupted data in the database: nodeId ' || nodeId || ': unkmown nodeType');
	END IF;
	
	RETURN result;

END getNodeValue;

/


CREATE OR REPLACE PROCEDURE setNodeValue (nodeId NUMBER, arg VARCHAR) AS
	nodeTypeId	NUMBER;
	CURSOR		cr_NodeType	IS SELECT NodeType FROM t_Node WHERE id = nodeId;
BEGIN

	OPEN	cr_NodeType;
	FETCH	cr_NodeType INTO nodeTypeId;
	IF cr_NodeType%NOTFOUND THEN
		raise_application_error (-20000, 'Node ' || nodeId || ' does not exist in the database');
	END IF;
	CLOSE	cr_NodeType;

	IF	nodeTypeId = 1	-- Element
	THEN	DOMException (7);
			
	ELSIF	nodeTypeId = 2	-- Attr
	THEN	UPDATE	t_Attr
		SET	value = arg
		WHERE	id = nodeId;
		
	ELSIF	nodeTypeId = 3	-- Text
	THEN	UPDATE	t_CharacterData
		SET	data = arg
		WHERE	id = nodeId;
	
	ELSIF	nodeTypeId = 4	-- CDATA
	THEN	UPDATE	t_CharacterData
		SET	data = arg
		WHERE	id = nodeId;
	
	ELSIF	nodeTypeId = 5	-- ENTITY_REFERENCE
	THEN	DOMException (7);
	
	ELSIF	nodeTypeId = 6	-- ENTITY
	THEN	DOMException (7);
	
	ELSIF	nodeTypeId = 7	-- PROCESSING_INSTRUCTION
	THEN	UPDATE	t_ProcessingInstruction
		SET	Data = arg
		WHERE	id = nodeId;
	
	ELSIF	nodeTypeId = 8	-- Comment
	THEN	UPDATE	t_CharacterData
		SET	data = arg
		WHERE	id = nodeId;
	
	ELSIF	nodeTypeId = 9	-- Document
	THEN	DOMException (7);
	
	ELSIF	nodeTypeId = 10	-- DocumentType
	THEN	DOMException (7);
	
	ELSIF	nodeTypeId = 11	-- DocumentFragment
	THEN	DOMException (7);
		
	ELSIF	nodeTypeId = 12	-- NOTATION
	THEN	DOMException (7);
	
	ELSE
		raise_application_error (-20000, 'internal error in getNodeName, corrupted data in the database: nodeId ' || nodeId || ': unkmown nodeType');
	END IF;
	
END setNodeValue;

/


-- readonly attribute  unsigned short       nodeType;
CREATE OR REPLACE FUNCTION getNodeType (NodeId NUMBER) RETURN NUMBER AS
	nodeTypeId	NUMBER;
	CURSOR		cr_NodeType	IS SELECT NodeType FROM t_Node WHERE id = nodeId;
BEGIN

	OPEN	cr_NodeType;
	FETCH	cr_NodeType INTO nodeTypeId;
	IF cr_NodeType%NOTFOUND THEN
		raise_application_error (-20000, 'Node ' || nodeId || ' does not exist in the database');
	END IF;
	CLOSE	cr_NodeType;

	RETURN	nodeTypeId;
END getNodeType;

/


-- readonly attribute  Node                 parentNode;
CREATE OR REPLACE FUNCTION getParentNode (nodeId NUMBER) RETURN NUMBER AS
	parentNodeId	NUMBER;
BEGIN
	SELECT	parent_node_id
	INTO	parentNodeId
	FROM	t_arch
	WHERE	specified_in_dom = 1 AND
		child_node_id = nodeId;
	-- only works with trees, as per the Rec
	-- if the Node has more than one parentNode (in a poset), the implicit cursor will raise an error
	-- once enough data is available, check this!
	
	RETURN	parentNodeId;
END getParentNode;

/


-- readonly attribute  NodeList             childNodes;
-- not implemented untill ResultSet issue is resolved



-- readonly attribute  Node                 firstChild;
CREATE OR REPLACE FUNCTION getFirstChild (nodeId NUMBER) RETURN NUMBER AS
	CURSOR	cr_childNodes IS
		SELECT	child_node_id
		FROM	t_arch
		WHERE	specified_in_dom = 1 AND
			parent_node_id = nodeId
		ORDER BY child_ord_no ASC;
	firstChildId	NUMBER;
BEGIN
	OPEN	cr_childNodes;
	FETCH	cr_childNodes INTO firstChildId;
	CLOSE	cr_childNodes;
	
	RETURN	firstChildId;
END getFirstChild;

/


-- readonly attribute  Node                 lastChild;
CREATE OR REPLACE FUNCTION getLastChild (nodeId NUMBER) RETURN NUMBER AS
	CURSOR	cr_childNodes IS
		SELECT	child_node_id
		FROM	t_arch
		WHERE	specified_in_dom = 1 AND
			parent_node_id = nodeId
		ORDER BY child_ord_no DESC;
	lastChildId	NUMBER;
BEGIN
	OPEN	cr_childNodes;
	FETCH	cr_childNodes INTO lastChildId;
	CLOSE	cr_childNodes;
	
	RETURN	lastChildId;
END getLastChild;

/


-- readonly attribute  Node                 previousSibling;
CREATE OR REPLACE FUNCTION getPreviousSibiling (nodeId NUMBER) RETURN NUMBER AS
	CURSOR	cr_prevSibilings IS
		SELECT	parent_arches.child_node_id
		FROM	t_arch my_arch,
			t_arch parent_arches
		WHERE	my_arch.parent_node_id = parent_arches.child_node_id AND
			my_arch.specified_in_dom = 1 AND
			parent_arches.specified_in_dom = 1 AND
			my_arch.child_node_id = nodeId AND
			parent_arches.child_ord_no < my_arch.child_ord_no
		ORDER BY parent_arches.child_ord_no DESC;
	previousSibilingId	NUMBER;
BEGIN
	-- if the Node has more than one parent, the results are unpredictable.
	-- consider raising an error if the Node has more one parent.
	OPEN	cr_prevSibilings;
	FETCH	cr_prevSibilings INTO previousSibilingId;
	CLOSE	cr_prevSibilings;

	RETURN previousSibilingId;
END getPreviousSibiling;

/


-- readonly attribute  Node                 nextSibling;
CREATE OR REPLACE FUNCTION getNextSibiling (nodeId NUMBER) RETURN NUMBER AS
	CURSOR	cr_nextSibilings IS
		SELECT	parent_arches.child_node_id
		FROM	t_arch my_arch,
			t_arch parent_arches
		WHERE	my_arch.parent_node_id = parent_arches.child_node_id AND
			my_arch.specified_in_dom = 1 AND
			parent_arches.specified_in_dom = 1 AND
			my_arch.child_node_id = nodeId AND
			parent_arches.child_ord_no > my_arch.child_ord_no
		ORDER BY parent_arches.child_ord_no ASC;
	nextSibilingId	NUMBER;
BEGIN
	-- if the Node has more than one parent, the results are unpredictable.
	-- consider raising an error if the Node has more one parent.
	OPEN	cr_nextSibilings;
	FETCH	cr_nextSibilings INTO nextSibilingId;
	CLOSE	cr_nextSibilings;

	RETURN nextSibilingId;
END getNextSibiling;

/


-- readonly attribute  NamedNodeMap         attributes;
-- not implemented untill ResultSet issue is resolved


-- readonly attribute  Document             ownerDocument;
CREATE OR REPLACE FUNCTION getOwnerDocument (nodeId NUMBER) RETURN NUMBER AS
	ownerDocumentId	NUMBER;
BEGIN
	-- slentlty returns a NULL if nodeId doesn't exist in the database
	SELECT	ownerDocument
	INTO	ownerDocumentId
	FROM	t_Node
	WHERE	id = nodeId;
	
	RETURN	ownerDocumentId;
END getOwnerDocument;

/
	

-- Node                      insertBefore(in Node newChild,
-- in Node refChild)
-- raises(DOMException);
CREATE OR REPLACE FUNCTION insertBefore (nodeId NUMBER, newChild NUMBER, refChild NUMBER) RETURN NUMBER AS
-- DocumnetFragment instanes are not handled yet; inserting a Document will raise HIERARCH_REQUEST_ERROR
	PRAGMA			AUTONOMOUS_TRANSACTION;		-- ?
	CURSOR	cr_archType	IS
		SELECT	t_archType.id
		FROM	t_archType,
			t_Node t_parent_node,
			t_Node t_child_node
		WHERE	t_archType.parent_node_type_id = t_parent_node.nodeType AND
			t_archType.child_node_type_id = t_child_node.nodeType AND
			t_parent_node.id = nodeId AND
			t_child_node.id = newChild AND
			-- Attribute Nodes cannot be inserted with this method
			specified_in_dom = 1;
	CURSOR	cr_ownerDocument	IS
	-- alternatively, NVL (getOwnerDocument (nodeId), nodeId) = getOwnerDocument(newChild) can be checked
		SELECT	parent.id
		FROM	t_Node parent,
			t_Node child
		WHERE	NVL (parent.ownerDocument, parent.id) = child.ownerDocument AND
			parent.id = nodeId AND
			child.id = newChild;
	CURSOR	cr_refChildArch	IS
		SELECT	child_ord_no
		FROM	t_arch
		WHERE	parent_node_id = nodeId AND
			child_node_id = refChild AND
			specified_in_dom = 1;
	newArchTypeId	NUMBER;
	dummy		NUMBER;
	refChildOrdNo	NUMBER;
BEGIN
	-- check that all the nodes exist
	
	-- HIERARCHY_REQUEST_ERR: Raised if this node is of a type that does not allow children of the type of the newChild node,
	--! or if the node to insert is one of this node's ancestors.
	OPEN	cr_archType;
	FETCH	cr_archType INTO newArchTypeId;
	IF	newArchTypeId IS NULL THEN
		DOMException (3);
	END IF;
	CLOSE	cr_archType;

	-- WRONG_DOCUMENT_ERR: Raised if newChild was created from a different document than the one that created this node.
	OPEN	cr_ownerDocument;
	FETCH	cr_ownerDocument INTO dummy;
	CLOSE	cr_ownerDocument;
	IF	dummy IS NULL THEN
		DOMException (4);
	END IF;

	-- NO_MODIFICATION_ALLOWED_ERR: Raised if this node is readonly.
	-- further research is needed

	-- NOT_FOUND_ERR: Raised if refChild is not a child of this node.
	OPEN	cr_refChildArch;
	FETCH	cr_refChildArch INTO refChildOrdNo;
	CLOSE	cr_refChildArch;
	IF	refChildOrdNo IS NULL THEN	-- child_ord_no is declared NOT NULL
						-- this SELECT is used for two purpuses:
		DOMException (8);
	END IF;
	
	-- after all the error checking, start the actual work
	
	-- It is important that DELETE precedes the UPDATE; consider moving sibilings around - a gap in numbering will occure
	
	DELETE
	FROM	t_arch
	WHERE	child_node_id = newChild;
        -- this will delete _all_ existing references to newChild; for posets, this will need to be rewritten

        UPDATE	t_arch
        SET	child_ord_no = child_ord_no + 1
        WHERE	parent_node_id = nodeId AND
        	child_ord_no <= refChildOrdNo;
	
	INSERT	INTO
	t_arch	(id, parent_node_id, child_node_id, parent_ord_no, child_ord_no, type_id, specified_in_dom)
        VALUES	(sq_arch.NEXTVAL, nodeId, newChild, 1, refChildOrdNo, newArchTypeId, 1);

        COMMIT;
	
	RETURN newChild;
END insertBefore;

/

	
-- Node                      replaceChild(in Node newChild,
-- in Node oldChild)
-- raises(DOMException);
CREATE OR REPLACE FUNCTION replaceChild (nodeId NUMBER, newChild NUMBER, oldChild NUMBER) RETURN NUMBER AS
-- DocumnetFragment instanes are not handled yet; inserting a Document will raise HIERARCH_REQUEST_ERROR
	PRAGMA			AUTONOMOUS_TRANSACTION;		-- ?
	CURSOR	cr_archType	IS
		SELECT	t_archType.id
		FROM	t_archType,
			t_Node t_parent_node,
			t_Node t_child_node
		WHERE	t_archType.parent_node_type_id = t_parent_node.nodeType AND
			t_archType.child_node_type_id = t_child_node.nodeType AND
			t_parent_node.id = nodeId AND
			t_child_node.id = newChild AND
			-- Attribute Nodes cannot be inserted with this method
			specified_in_dom = 1;
	CURSOR	cr_ownerDocument	IS
	-- alternatively, NVL (getOwnerDocument (nodeId), nodeId) = getOwnerDocument(newChild) can be checked
		SELECT	parent.id
		FROM	t_Node parent,
			t_Node child
		WHERE	NVL (parent.ownerDocument, parent.id) = child.ownerDocument AND
			parent.id = nodeId AND
			child.id = newChild;
	CURSOR	cr_oldChildArch	IS
		SELECT	child_ord_no
		FROM	t_arch
		WHERE	parent_node_id = nodeId AND
			child_node_id = oldChild AND
			specified_in_dom = 1;
	newArchTypeId	NUMBER;
	dummy		NUMBER;
	oldChildOrdNo	NUMBER;
BEGIN
	-- check that all the nodes exist
	
	-- HIERARCHY_REQUEST_ERR: Raised if this node is of a type that does not allow children of the type of the newChild node,
	--! or if the node to insert is one of this node's ancestors.
	OPEN	cr_archType;
	FETCH	cr_archType INTO newArchTypeId;
	IF	newArchTypeId IS NULL THEN
		DOMException (3);
	END IF;
	CLOSE	cr_archType;

	-- WRONG_DOCUMENT_ERR: Raised if newChild was created from a different document than the one that created this node.
	OPEN	cr_ownerDocument;
	FETCH	cr_ownerDocument INTO dummy;
	CLOSE	cr_ownerDocument;
	IF	dummy IS NULL THEN
		DOMException (4);
	END IF;

	-- NO_MODIFICATION_ALLOWED_ERR: Raised if this node is readonly.
	-- further research is needed

	-- NOT_FOUND_ERR: Raised if oldChild is not a child of this node.
	OPEN	cr_oldChildArch;
	FETCH	cr_oldChildArch INTO oldChildOrdNo;
	CLOSE	cr_oldChildArch;
	IF	oldChildOrdNo IS NULL THEN	-- child_ord_no is declared NOT NULL
						-- this SELECT is used for two purpuses:
		DOMException (8);
	END IF;
	
	-- after all the error checking, start the actual work
	
	DELETE
	FROM	t_arch
	WHERE	child_node_id = newChild AND
		child_node_id != oldChild;	-- for the case when newChild = oldChild; altenative test: newChild != oldChild. if this is the first condition, Oracle may skip the DELETE altogether
        -- this will delete _all_ existing references to newChild; for posets, this will need to be rewritten


        UPDATE	t_arch
        SET	child_node_id = newChild,
        	type_id = newArchTypeId
        WHERE	parent_node_id = nodeId AND
        	child_node_id = oldChild;

	COMMIT;
	
	RETURN newChild;
END replaceChild;

/

	
	
	


--------------------------------------------------------------------------
-- Interface NodeList
--------------------------------------------------------------------------
-- Node                      item(in unsigned long index);
-- readonly attribute  unsigned long        length;


--------------------------------------------------------------------------
-- Interface NamedNodeMap
--------------------------------------------------------------------------
-- Node                      getNamedItem(in DOMString name);
-- Node                      setNamedItem(in Node arg)
-- raises(DOMException);
-- Node                      removeNamedItem(in DOMString name)
-- raises(DOMException);
-- Node                      item(in unsigned long index);
-- readonly attribute  unsigned long        length;


--------------------------------------------------------------------------
-- Interface CharacterData
--------------------------------------------------------------------------
-- attribute  DOMString            data;
-- // raises(DOMException) on setting
-- // raises(DOMException) on retrieval
-- readonly attribute  unsigned long        length;
-- DOMString                 substringData(in unsigned long offset,
-- in unsigned long count)
-- raises(DOMException);
-- void                      appendData(in DOMString arg)
-- raises(DOMException);
-- void                      insertData(in unsigned long offset,
-- in DOMString arg)
-- raises(DOMException);
-- void                      deleteData(in unsigned long offset,
-- in unsigned long count)
-- raises(DOMException);
-- void                      replaceData(in unsigned long offset,
-- in unsigned long count,
-- in DOMString arg)
-- raises(DOMException);


--------------------------------------------------------------------------
-- Interface Attr
--------------------------------------------------------------------------
-- readonly attribute  DOMString            name;
-- readonly attribute  boolean              specified;
-- attribute  DOMString            value;


--------------------------------------------------------------------------
-- Interface Element
--------------------------------------------------------------------------
-- readonly attribute  DOMString            tagName;
-- DOMString                 getAttribute(in DOMString name);
-- void                      setAttribute(in DOMString name,
-- in DOMString value)
-- raises(DOMException);
-- void                      removeAttribute(in DOMString name)
-- raises(DOMException);
-- Attr                      getAttributeNode(in DOMString name);
-- Attr                      setAttributeNode(in Attr newAttr)
-- raises(DOMException);
-- Attr                      removeAttributeNode(in Attr oldAttr)
-- raises(DOMException);
-- NodeList                  getElementsByTagName(in DOMString name);
-- void                      normalize();


--------------------------------------------------------------------------
-- Interface Text
--------------------------------------------------------------------------
-- Text                      splitText(in unsigned long offset)
-- raises(DOMException);


--------------------------------------------------------------------------
-- Interface Comment
--------------------------------------------------------------------------


--------------------------------------------------------------------------
-- Interface CDATASection
--------------------------------------------------------------------------


--------------------------------------------------------------------------
-- Interface DocumentType
--------------------------------------------------------------------------
-- readonly attribute  DOMString            name;
-- readonly attribute  NamedNodeMap         entities;
-- readonly attribute  NamedNodeMap         notations;


--------------------------------------------------------------------------
-- Interface Notation
--------------------------------------------------------------------------
-- readonly attribute  DOMString            publicId;
-- readonly attribute  DOMString            systemId;


--------------------------------------------------------------------------
-- Interface Entity
--------------------------------------------------------------------------
-- readonly attribute  DOMString            publicId;
-- readonly attribute  DOMString            systemId;
-- readonly attribute  DOMString            notationName;


--------------------------------------------------------------------------
-- Interface EntityReference
--------------------------------------------------------------------------


--------------------------------------------------------------------------
-- Interface ProcessingInstruction
--------------------------------------------------------------------------
-- readonly attribute  DOMString            target;
-- attribute  DOMString            data;
-- // raises(DOMException) on setting


