--------------------------------------------------------------------------
-- Private functions
--------------------------------------------------------------------------
CREATE	FUNCTION nodeExists	(INT)	RETURNS BOOLEAN AS '
	SELECT	COUNT (id) > 0
	FROM	t_node
	WHERE	id = $1;
' LANGUAGE 'sql';

CREATE	FUNCTION nodeTypeMatches	(INT, SMALLINT)	RETURNS BOOLEAN AS '
	SELECT	NodeType = $2
	FROM	t_node
	WHERE	id = $1;
' LANGUAGE 'sql';

-- checks if a string can be inserted. Caller should raise INVALID_CHARACTER_ERR DOMException is true.
-- currently not implemented, a stub only; see ISSUES
CREATE	FUNCTION	containsInvalidCharacters (VARCHAR)	RETURNS BOOLEAN AS '
	SELECT	FALSE;
' LANGUAGE 'sql';


--------------------------------------------------------------------------
-- Proprietary functions
--------------------------------------------------------------------------


--------------------------------------------------------------------------
-- Exception DOMException
--------------------------------------------------------------------------
CREATE	FUNCTION DOMException	(SMALLINT) RETURNS INT AS '
    DECLARE	excMsg VARCHAR;
    -- the prefix to be added to the message of a DOMEXception to let the
    -- application know that a DOMException was thrown, as opposed to a regular exception
    DECLARE	excMsgPrefix VARCHAR;
    BEGIN
    	SELECT INTO excMsg msg FROM t_DOMException WHERE id = $1;
    	SELECT INTO excMsgPrefix value FROM t_settings WHERE name = ''domExceptionPrefix'';
    	IF excMsg IS NULL
    	    THEN
    		RAISE EXCEPTION ''Error rising DOMException, invalid exception code %'', $1;
    	    ELSE
	    	RAISE EXCEPTION ''% % %'', excMsgPrefix, $1, excMsg;
	END IF;
        RETURN $1;
    END;	-- DOMException
' LANGUAGE 'plpgsql';


--------------------------------------------------------------------------
-- Interface DOMImplementation
--------------------------------------------------------------------------
CREATE	FUNCTION hasFeature	(VARCHAR, VARCHAR)	RETURNS BOOLEAN
AS '
	SELECT	COUNT (*) > 0
	AS	hasFeature
	FROM	t_DOMImplementation
	WHERE	UPPER (feature) = UPPER ($1) AND
		UPPER (version) = COALESCE ($2, version);
		-- This is the version number of the package name to test.
		-- In Level 1, this is the string "1.0". If the version is not specified,
		-- supporting any version of the feature will cause the method to return true.
		-- (Rec)
		--
		-- if $2 (version) is NULL, t_DOMException.version is compared to itself,
		-- effectively ignoring the test
' LANGUAGE 'sql';


-- creates an empty, DTD-less document
-- unspecified in DOM level 1; propriatory extention
CREATE	FUNCTION	createNamedDocument (VARCHAR)	RETURNS INT AS '
    DECLARE
    	DOCUMENT_NODE_TYPE_ID	CONSTANT INT = 9;
    	newDocumentId			 INT;
    	documentName			 ALIAS FOR $1;
    BEGIN
	-- create the Node part
	newDocumentId := createNode (NULL, DOCUMENT_NODE_TYPE_ID);
	
	INSERT	INTO t_Document (id, name)
	VALUES	(newDocumentId, documentName);
	
	RETURN	newDocumentId;
    END;	-- createNewDocument
' LANGUAGE 'plpgsql';

--------------------------------------------------------------------------
-- Non-DOM functions for access to Document nodes stored in this database
-- Documents are accessible through a read-only NamedNodeMap in DBDOMImplementation
--------------------------------------------------------------------------

CREATE	FUNCTION getNofDocuments()	RETURNS INT
AS '
	SELECT  COUNT (id)
	FROM    t_Document;
' LANGUAGE 'sql';

CREATE	FUNCTION getDocumentByOrdNo (INT)	RETURNS INT
AS '
	SELECT  id
	FROM    t_Document as t_this_document
	WHERE   (SELECT  COUNT (t_preceding_documents.id)
	         FROM    t_Document as t_preceding_documents
	         WHERE   t_preceding_documents.id < t_this_document.id) = $1;
' LANGUAGE 'sql';

CREATE	FUNCTION getDocumentByName (VARCHAR)	RETURNS INT
AS '
	SELECT  id
	FROM    t_Document
	WHERE   name = $1;
' LANGUAGE 'sql';

CREATE	FUNCTION getDocumentById   (INT)	RETURNS INT
AS '
	SELECT  id
	FROM    t_Document
	WHERE   id = $1;
' LANGUAGE 'sql';
	


--------------------------------------------------------------------------
-- Interface DocumentFragment
--------------------------------------------------------------------------


--------------------------------------------------------------------------
-- Interface Document
--------------------------------------------------------------------------

-- readonly attribute  DocumentType         doctype;
-- Currently, DTDs aer not handled in DBDOM, so DocumentType is allways NULL
-- 31.8.2000: not true. DOcumentType has three read-only attributes that can be accessed. change as soon as possible. see oracle version for correct implementation.
CREATE	FUNCTION getDocType (INT) RETURNS INT
AS '
    BEGIN
	RETURN NULL;
    END;
' LANGUAGE 'plpgsql';


-- readonly attribute  DOMImplementation    implementation;
-- "The DOMImplementation interface provides a number of methods for performing operations that are independent of any particular instance of the document object model."
-- The only method DOMImplementation has in Level 1 is hasFeature. There is no object to reference in the database.


-- readonly attribute  Element              documentElement;
-- this function hasn't been tested yet for the lack of data
CREATE	FUNCTION getDocumentElement	(INT) RETURNS INT
AS '
-- consider checking count(id); it should be 0 for an empty Document, and exactly 1 otherwise
	SELECT	t_arch.child_node_id
	FROM	t_arch,
		t_arch_type
	WHERE	t_arch.type_id = t_arch_type.id AND
		parent_node_id = $1 AND
		t_arch_type.child_node_type_id = 1;
' LANGUAGE 'sql';


-- private funcion to create the t_Node part of a new node
-- this function performs no validation; it should never be called directly
-- this function will likely differ between DB platforms, since getting
-- the uid of a newly inserted record is not defined by SQL
CREATE	FUNCTION createNode	(INT, SMALLINT)	RETURNS INT
AS '
	INSERT	INTO t_Node (NodeType, ownerDocument)
	VALUES	($2, $1);

	-- poratble and simple, but not very robust	
	SELECT	MAX (id)
	FROM	t_Node;	
' LANGUAGE 'sql';


-- this function checks that the parent node exists and that it is a Document or a DocumentFragment.
-- wrong: DocumentFragment cannot create Nodes
CREATE	FUNCTION parentNodeIsValid	(INT) RETURNS	BOOLEAN
AS '
    DECLARE	parentDocumentId	ALIAS FOR $1;
    BEGIN
    	-- check that parentDocument exists
    	IF NOT nodeExists (parentDocumentId) THEN
    		RAISE EXCEPTION ''Node doesn''''t exist'';
    	END IF;
    		
    	-- check that parentDocument is a Document or DocumentFragment
    	IF NOT nodeTypeMatches (parentDocumentId, 9) AND
    	   NOT nodeTypeMatches (parentDocumentId, 11) THEN
    	   	RAISE EXCEPTION ''Type mismatch: expected Document'';
    	END IF;
    	
    	-- if evereything is valid,
    	RETURN	TRUE;
    END;	-- parentNodeIsValid
' LANGUAGE 'plpgsql';


-- Element                   createElement(in DOMString tagName)
-- raises(DOMException);
CREATE	FUNCTION createElement	(INT, VARCHAR)	RETURNS INT
AS '
    DECLARE	parentDocumentId	ALIAS FOR $1;
		tagName			ALIAS FOR $2;
		tagId			INT;
		newElementId		INT;
    BEGIN
    	-- check that parentDocument is a valid existing Document
    	IF parentNodeIsValid (parentDocumentId)
    	    THEN
	    	-- check that tagName does not contain invalid characters
	    	IF containsInvalidCharacters (tagName)	THEN
	    		PERFORM DOMException (5);	-- INVALID_CHARACTER_ERR
		END IF;
    	   	
    		-- get tag id
		SELECT INTO tagId id FROM t_tag WHERE NAME = tagName;
		IF tagId IS NULL THEN
		    BEGIN
		    	INSERT INTO t_tag (name) VALUES (tagName);
	    		SELECT INTO tagId id FROM t_tag WHERE NAME = tagName;	-- not very efficient, but robust
		    END;		
		END IF;
	
		-- create new Element
		-- create the Node part
		newElementId := createNode (parentDocumentId, 1);
	
		-- create the Element part
		INSERT	INTO t_Element	(id, tag_id)
		VALUES	(newElementId, tagId);
			
		-- return a reference to the new Element
		RETURN newElementId;
	END IF;
    END;	-- createElement
' LANGUAGE 'plpgsql';



-- DocumentFragment          createDocumentFragment();
CREATE	FUNCTION createDocumentFragment	(INT)	RETURNS INT
AS '
    DECLARE	parentDocumentId	ALIAS FOR $1;
    		newDocumentFragmentId	INT;
    BEGIN
    	-- check that parentDocument is a valid existing Document
    	IF parentNodeIsValid (parentDocumentId)
    	    THEN    	   	
		-- create new DocumentFragment
		-- create the Node part
		newDocumentFragmentId := createNode (parentDocumentId, 11);

		-- create the DocumentFragment part
		INSERT	INTO t_DocumentFragment	(id)
		VALUES	(newDocumentFragmentId);
	
		--return a reference to the new DocumentFragment
		RETURN	newDocumentFragmentId;
	END IF;
    END;	-- createDocumentFragment
' LANGUAGE 'plpgsql';


-- a private generic function to create a CharacterData node
-- createTextNode, createComment and createCDATASection call this function
CREATE	FUNCTION createCharacterDataNode	(INT, VARCHAR, SMALLINT)	RETURNS INT
AS '
    DECLARE	parentDocumentId	ALIAS FOR $1;
    		textData		ALIAS FOR $2;
    		nodeTypeId		ALIAS FOR $3;
    		newCharacterDataNodeId	INT;
    BEGIN
    	-- check that parentDocument is a valid existing Document
    	IF parentNodeIsValid (parentDocumentId)
    	    THEN
	    	-- check that nodeTypeId is TextNode (3), Comment (8) or CDATASection (4)
    		IF  NOT nodeTypeId = 3 AND
	    	    NOT nodeTypeId = 8 AND
		    NOT nodeTypeId = 4 THEN
			RAISE EXCEPTION ''Invalid use of private function: createCharacterDataNode can only create TextNode, Comment or CDATASection; % passed as nodeTypeId'', nodeTypeId;
	    	END IF;
	    	
		-- create new CaracterData node
		-- create the Node part
		newCharacterDataNodeId := createNode (parentDocumentId, nodeTypeId);
	
		-- create the CharacterData part
		INSERT	INTO t_CharacterData	(id, data)
		VALUES	(newCharacterDataNodeId, textData);
	
		--return a reference to the new TextNode
		RETURN	newCharacterDataNodeId;
	END IF;
    END;	-- createCharacterDataNode
' LANGUAGE 'plpgsql';


-- Text                      createTextNode(in DOMString data);
CREATE	FUNCTION createTextNode	(INT, VARCHAR)	RETURNS INT
AS '
    DECLARE	parentDocumentId	ALIAS FOR $1;
    		textData		ALIAS FOR $2;
    		newTextNodeId		INT;
    BEGIN
    	RETURN	createCharacterDataNode (parentDocumentId, textData, 3);
    END;
' LANGUAGE 'plpgsql';


-- Comment                   createComment(in DOMString data);
CREATE	FUNCTION createComment	(INT, VARCHAR)	RETURNS INT
AS '
    DECLARE	parentDocumentId	ALIAS FOR $1;
    		textData		ALIAS FOR $2;
    		newCommentNodeId	INT;
    BEGIN
    	RETURN	createCharacterDataNode (parentDocumentId, textData, 8);
    END;	-- createComment
' LANGUAGE 'plpgsql';

-- CDATASection              createCDATASection(in DOMString data)
-- raises(DOMException);  // NOT_SUPPORTED_ERR: Raised if this document is an HTML document.
-- Since DBDOM doesn't supprt HTML, this exception is never raised
CREATE FUNCTION createCDATASection	(INT, VARCHAR)	RETURNS INT
AS '
    DECLARE	parentDocumentId	ALIAS FOR $1;
    		textData		ALIAS FOR $2;
    		newCDATASectionNodeId	INT;
    BEGIN
    	RETURN	createCharacterDataNode (parentDocumentId, textData, 4);
    END;	-- createCDATASection
' LANGUAGE 'plpgsql';


-- ProcessingInstruction     createProcessingInstruction(in DOMString target,
-- in DOMString data)
-- raises(DOMException);
CREATE	FUNCTION createProcessingInstruction	(INT, VARCHAR, VARCHAR)	RETURNS INT
AS '
    DECLARE	parentDocumentId	ALIAS FOR $1;
    		targetStr		ALIAS FOR $2;
    		dataStr			ALIAS FOR $3;
    		newProcessingInstructionId	INT;
    BEGIN
    	-- check that parentDocument is a valid existing Document
    	IF parentNodeIsValid (parentDocumentId)
    	    THEN
		-- create new CaracterData node
		-- create the Node part
		newProcessingInstructionId := createNode (parentDocumentId, 7);
		
		-- create the ProcessingInstruction part
		INSERT	INTO t_ProcessingInstruction	(id, target, data)
		VALUES	(newProcessingInstructionId, targetStr, dataStr);
		--return a reference to the new TextNode
		RETURN	newProcessingInstructionId;
	END IF;
    END;	-- createProcessingInstruction
' LANGUAGE 'plpgsql';


-- Attr                      createAttribute(in DOMString name)
-- raises(DOMException);
CREATE	FUNCTION createAttribute	(INT, VARCHAR)	RETURNS INT
AS '
    DECLARE	parentDocumentId	ALIAS FOR $1;
		attrName		ALIAS FOR $2;
		attrNameId		INT;
		newAttributeId		INT;
    BEGIN
    	-- check that parentDocument is a valid existing Document
    	IF parentNodeIsValid (parentDocumentId)
    	    THEN
	    	-- check that tagName does not contain invalid characters
	    	IF containsInvalidCharacters (attrName)	THEN
	    		PERFORM DOMException (5);	-- INVALID_CHARACTER_ERR
		END IF;
    	   	
    		-- get tag id
		SELECT INTO attrNameId id FROM t_attr_name WHERE NAME = attrName;
		IF attrNameId IS NULL THEN
		    BEGIN
		    	INSERT INTO t_attr_name (name) VALUES (attrName);
	    		SELECT INTO attrNameId id FROM t_attr_name WHERE NAME = attrName;	-- not very efficient, but robust
		    END;		
		END IF;
	
		-- create new Element
		-- create the Node part
		newAttributeId := createNode (parentDocumentId, 2);
	
		-- create the Element part
		INSERT	INTO t_Attr	(id, name_id, specified)
		VALUES	(newAttributeId, attrNameId, TRUE);
			
		-- return a reference to the new Element
		RETURN newAttributeId;
	END IF;
    END;	-- createAttribute
' LANGUAGE 'plpgsql';


-- EntityReference           createEntityReference(in DOMString name)
-- raises(DOMException);
CREATE	FUNCTION createEntityReference(INT, VARCHAR) RETURNS	INT
AS '
    DECLARE	parentDocumentId	ALIAS FOR $1;
    		nameStr			ALIAS FOR $2;
    BEGIN
	RAISE EXCEPTION ''createEntityReference() is not implemented yet'';
    END;	-- createEntityReference
' LANGUAGE 'plpgsql';

-- NodeList                  getElementsByTagName(in DOMString tagname);

-- appearently, this it the same method as Element.getElementsByTagName

--------------------------------------------------------------------------
-- non-DOM functions
CREATE FUNCTION getDocumentName	(INT)	RETURNS VARCHAR
AS '
	SELECT  name
	FROM    t_Document
	WHERE   id = $1;
' LANGUAGE 'sql';


--------------------------------------------------------------------------
-- Interface Node
--------------------------------------------------------------------------
-- readonly attribute  DOMString            nodeName;
CREATE	FUNCTION getNodeName	(INT) RETURNS VARCHAR
AS '
    DECLARE	nodeId		ALIAS FOR $1;
		nodeTypeId	INT;
		result		VARCHAR;
		resultFound	BOOLEAN := FALSE;
    BEGIN
	SELECT	INTO
	nodeTypeId	NodeType
	FROM	t_node
	WHERE	id = nodeId;

	IF	NOT resultFound AND
		nodeTypeId = 1	-- Element
	THEN	SELECT	INTO
		result	t_tag.name
		FROM	t_Element,
			t_tag
		WHERE	t_Element.tag_id = t_tag.id AND
			t_Element.id     = nodeId;
		resultFound := TRUE;
	END IF;
			
	IF	NOT resultFound AND
		nodeTypeId = 2	-- Attr
	THEN	SELECT	INTO
		result	t_attr_name.name
		FROM	t_Attr,
			t_attr_name
		WHERE	t_Attr.name_id = t_attr_name.id AND
			t_Attr.id = nodeId;
		resultFound := TRUE;
	END IF;
		
	IF	NOT resultFound AND
		nodeTypeId = 3	-- Text
	THEN	result := ''#text'';
		resultFound := TRUE;
	END IF;
	
	IF	NOT resultFound AND
		nodeTypeId = 4	-- CDATA
	THEN	result := ''#cdata-section'';
		resultFound := TRUE;
	END IF;
	
	IF	NOT resultFound AND
		nodeTypeId = 5	-- ENTITY_REFERENCE
	THEN	RAISE EXCEPTION ''entity references are not handled yet'';
		resultFound := TRUE;
	END IF;
	
	IF	NOT resultFound AND
		nodeTypeId = 6	-- ENTITY
	THEN	SELECT	INTO
		result	name
		FROM	t_Entity
		WHERE	id = nodeId;
		resultFound := TRUE;
	END IF;
	
	IF	NOT resultFound AND
		nodeTypeId = 7	-- PROCESSING_INSTRUCTION
	THEN	SELECT	INTO
		result	target
		FROM	t_ProcessingInstruction
		WHERE	id = nodeId;
		resultFound := TRUE;
	END IF;
	
	IF	NOT resultFound AND
		nodeTypeId = 8	-- Comment
	THEN	result := ''#comment'';
		resultFound := TRUE;
	END IF;
	
	IF	NOT resultFound AND
		nodeTypeId = 9	-- Document
	THEN	result := ''#document'';
		resultFound := TRUE;
	END IF;
	
	IF	NOT resultFound AND
		nodeTypeId = 10	-- DocumentType
	THEN	SELECT	INTO
		result	name
		FROM	t_Document_type
		WHERE	id = nodeId;
		resultFound := TRUE;
	END IF;
	
	IF	NOT resultFound AND
		nodeTypeId = 11	-- DocumentFragment
	THEN	result := ''#document-fragment'';
		resultFound := TRUE;
	END IF;
		
	IF	NOT resultFound AND
		nodeTypeId = 12	-- NOTATION
	THEN	SELECT	INTO
		result	name
		FROM	t_Notation
		WHERE	id = nodeId;
		resultFound := TRUE;
	END IF;
	
	IF resultFound
	    THEN
		RETURN result;
	    ELSE
		RAISE EXCEPTION ''internal error in getNodeName, nodeId %: unkmown nodeType'', nodeId;
	END IF;

    END;	-- getNodeName
' LANGUAGE 'plpgsql';



-- attribute  DOMString            nodeValue;
-- // raises(DOMException) on setting
-- // raises(DOMException) on retrieval
CREATE	FUNCTION getNodeValue (INT) RETURNS VARCHAR
AS '
    DECLARE	nodeId		ALIAS FOR $1;
		nodeTypeId	INT;
		result		VARCHAR;
		resultFound	BOOLEAN := FALSE;
    BEGIN
	SELECT	INTO
	nodeTypeId	NodeType
	FROM	t_node
	WHERE	id = nodeId;

	IF	NOT resultFound AND
		nodeTypeId = 1	-- Element
	THEN	result := NULL;
		resultFound := TRUE;
	END IF;
			
	IF	NOT resultFound AND
		nodeTypeId = 2	-- Attr
	THEN	SELECT	INTO
		result	value
		FROM	t_Attr
		WHERE	id = nodeId;
		resultFound := TRUE;
	END IF;
		
	IF	NOT resultFound AND
		nodeTypeId = 3	-- Text
	THEN	SELECT	INTO
		result	data
		FROM	t_CharacterData
		WHERE	id = nodeId;
		resultFound := TRUE;
	END IF;
	
	IF	NOT resultFound AND
		nodeTypeId = 4	-- CDATA
	THEN	SELECT	INTO
		result	data
		FROM	t_CharacterData
		WHERE	id = nodeId;
		resultFound := TRUE;
	END IF;
	
	IF	NOT resultFound AND
		nodeTypeId = 5	-- ENTITY_REFERENCE
	THEN	result := NULL;
		resultFound := TRUE;
	END IF;
	
	IF	NOT resultFound AND
		nodeTypeId = 6	-- ENTITY
	THEN	result := NULL;
		resultFound := TRUE;
	END IF;
	
	IF	NOT resultFound AND
		nodeTypeId = 7	-- PROCESSING_INSTRUCTION
	THEN	SELECT	INTO
		result	Data
		FROM	t_ProcessingInstruction
		WHERE	id = nodeId;
		resultFound := TRUE;
	END IF;
	
	IF	NOT resultFound AND
		nodeTypeId = 8	-- Comment
	THEN	SELECT	INTO
		result	data
		FROM	t_CharacterData
		WHERE	id = nodeId;
		resultFound := TRUE;
	END IF;
	
	IF	NOT resultFound AND
		nodeTypeId = 9	-- Document
	THEN	result := NULL;
		resultFound := TRUE;
	END IF;
	
	IF	NOT resultFound AND
		nodeTypeId = 10	-- DocumentType
	THEN	result := NULL;
		resultFound := TRUE;
	END IF;
	
	IF	NOT resultFound AND
		nodeTypeId = 11	-- DocumentFragment
	THEN	result := NULL;
		resultFound := TRUE;
	END IF;
		
	IF	NOT resultFound AND
		nodeTypeId = 12	-- NOTATION
	THEN	result := NULL;
		resultFound := TRUE;
	END IF;
	
	IF resultFound
	    THEN
		RETURN result;
	    ELSE
		RAISE EXCEPTION ''internal error in getNodeName, nodeId %: unkmown nodeType'', nodeId;
	END IF;

    END;	-- getNodeValue
' LANGUAGE 'plpgsql';


-- this function should have been void; return type is ignorable
CREATE	FUNCTION setNodeValue (INT, VARCHAR) RETURNS BOOLEAN
AS '
    DECLARE	nodeId		ALIAS FOR $1;
    		valueStr	ALIAS FOR $2;
		nodeTypeId	INT;
		valueSet	BOOLEAN := FALSE;
    BEGIN
	SELECT	INTO
	nodeTypeId	NodeType
	FROM	t_node
	WHERE	id = nodeId;

	IF	NOT valueSet AND
		nodeTypeId = 1	-- Element
	THEN	SELECT	DOMException (7);
		valueSet := TRUE;
	END IF;
			
	IF	NOT valueSet AND
		nodeTypeId = 2	-- Attr
	THEN	UPDATE	t_Attr
		SET	value = valueStr
		WHERE	id = nodeId;
		valueSet := TRUE;
	END IF;
		
	IF	NOT valueSet AND
		nodeTypeId = 3	-- Text
	THEN	UPDATE	t_CharacterData
		SET	data = valueStr
		WHERE	id = nodeId;
		valueSet := TRUE;
	END IF;
	
	IF	NOT valueSet AND
		nodeTypeId = 4	-- CDATA
	THEN	UPDATE	t_CharacterData
		SET	data = valueStr
		WHERE	id = nodeId;
		valueSet := TRUE;
	END IF;
	
	IF	NOT valueSet AND
		nodeTypeId = 5	-- ENTITY_REFERENCE
	THEN	SELECT	DOMException (7);
		valueSet := TRUE;
	END IF;
	
	IF	NOT valueSet AND
		nodeTypeId = 6	-- ENTITY
	THEN	SELECT	DOMException (7);
		valueSet := TRUE;
	END IF;
	
	IF	NOT valueSet AND
		nodeTypeId = 7	-- PROCESSING_INSTRUCTION
	THEN	UPDATE	t_ProcessingInstruction
		SET	data = valueStr
		WHERE	id = nodeId;
		valueSet := TRUE;
	END IF;
	
	IF	NOT valueSet AND
		nodeTypeId = 8	-- Comment
	THEN	UPDATE	t_CharacterData
		SET	data = valueStr
		WHERE	id = nodeId;
		valueSet := TRUE;
	END IF;
	
	IF	NOT valueSet AND
		nodeTypeId = 9	-- Document
	THEN	SELECT	DOMException (7);
		valueSet := TRUE;
	END IF;
	
	IF	NOT valueSet AND
		nodeTypeId = 10	-- DocumentType
	THEN	SELECT	DOMException (7);
		valueSet := TRUE;
	END IF;
	
	IF	NOT valueSet AND
		nodeTypeId = 11	-- DocumentFragment
	THEN	SELECT	DOMException (7);
		valueSet := TRUE;
	END IF;
		
	IF	NOT valueSet AND
		nodeTypeId = 12	-- NOTATION
	THEN	SELECT	DOMException (7);
		valueSet := TRUE;
	END IF;
	
	IF valueSet
	    THEN
	        -- consider returning null
		RETURN valueSet;
	    ELSE
		RAISE EXCEPTION ''internal error in getNodeName, nodeId %: unkmown nodeType'', nodeId;
	END IF;

    END;	-- setNodeValue
' LANGUAGE 'plpgsql';


-- readonly attribute  unsigned short       nodeType;
CREATE	FUNCTION getNodeType (INT) RETURNS SMALLINT
AS '
	SELECT	nodeType
	FROM	t_node
	WHERE	id = $1;
' LANGUAGE 'sql';

-- readonly attribute  Node                 parentNode;
CREATE	FUNCTION getParentNode (INT)	RETURNS INT
AS '
	SELECT	parent_node_id
	FROM	t_arch
	WHERE	specified_in_dom = TRUE AND
		child_node_id = $1;
' LANGUAGE 'sql';


-- readonly attribute  NodeList             childNodes;
CREATE	FUNCTION getChildNodes (INT)	RETURNS setof INT
AS '
	SELECT	child_node_id
	FROM	t_arch
	WHERE	specified_in_dom = TRUE AND
		parent_node_id = $1;
	;--ORDER BY t_arch.child_ord_no;	-- Postgres trows an error if ORDER BY is present
' LANGUAGE 'sql';


-- readonly attribute  Node                 firstChild;
CREATE	FUNCTION getFirstChild (INT)	RETURNS INT
AS '
	SELECT	child_node_id
	FROM	t_arch
	WHERE	specified_in_dom = TRUE AND
		parent_node_id = $1
	--ORDER BY child_ord_no ASC
	;--LIMIT	1;
' LANGUAGE 'sql';

-- readonly attribute  Node                 lastChild;
CREATE	FUNCTION getLastChild (INT)	RETURNS INT
AS '
	SELECT	child_node_id
	FROM	t_arch
	WHERE	specified_in_dom = TRUE AND
		parent_node_id = $1
	--ORDER BY child_ord_no DESC
	;--LIMIT	1;
' LANGUAGE 'sql';

-- readonly attribute  Node                 previousSibling;
CREATE	FUNCTION getPreviousSibling (INT)	RETURNS INT
AS '
	SELECT	sibiling_arch.child_node_id
	FROM	t_arch as my_arch,
		t_arch as sibiling_arch
	WHERE	my_arch.parent_node_id = sibiling_arch.parent_node_id AND
		my_arch.specified_in_dom = TRUE AND
		sibiling_arch.specified_in_dom = TRUE AND
		my_arch.child_node_id = $1 AND
		my_arch.child_ord_no > sibiling_arch.child_ord_no
	--OEDER BY sibiling_arch.child_ord_no DESC
	;--LIMIT	1;
' LANGUAGE 'sql';

-- readonly attribute  Node                 nextSibling;
CREATE	FUNCTION getNextSibling (INT)	RETURNS INT
AS '
	SELECT	sibiling_arch.child_node_id
	FROM	t_arch as my_arch,
		t_arch as sibiling_arch
	WHERE	my_arch.parent_node_id = sibiling_arch.parent_node_id AND
		my_arch.specified_in_dom = TRUE AND
		sibiling_arch.specified_in_dom = TRUE AND
		my_arch.child_node_id = $1 AND
		my_arch.child_ord_no < sibiling_arch.child_ord_no
	--OEDER BY sibiling_arch.child_ord_no ASC
	;--LIMIT	1;
' LANGUAGE 'sql';

-- readonly attribute  NamedNodeMap         attributes;
CREATE	FUNCTION getAttributes (INT)	RETURNS setof INT
AS '
	SELECT	t_arch.child_node_id
	FROM	t_arch,
		t_arch_type
	WHERE	t_arch.type_id = t_arch_type.id AND
		t_arch.parent_node_id = $1 AND
		t_arch_type.child_node_type_id = 2;	-- ATTRIBUTE_NODE
' LANGUAGE 'sql';

-- readonly attribute  Document             ownerDocument;
CREATE	FUNCTION getOwnerDocument (INT)	RETURNS	INT
AS '
	SELECT	ownerDocument
	FROM	t_Node
	WHERE	id = $1;
' LANGUAGE 'sql';

-- Node                      insertBefore(in Node newChild,
-- in Node refChild)
-- raises(DOMException);
CREATE	FUNCTION insertBefore (INT, INT, INT)	RETURNS INT
AS '
    DECLARE	thisNode	ALIAS FOR $1;
		newChild	ALIAS FOR $2;
    		refChild	ALIAS FOR $3;
    		thisNodeType	SMALLINT;
    		newChildType	SMALLINT;
    		refChildOrdNo	INT;
    		newArchType	INT := NULL;
    BEGIN
	-- HIERARCHY_REQUEST_ERR: Raised if this node is of a type that does not allow children of the type of the newChild node, or if the node to insert is one of this node''s ancestors.
	SELECT	INTO
	thisNodeType	NodeType
	FROM	t_node
	WHERE	id = thisNode;
	
	SELECT	INTO
	newChildType	NodeType
	FROM	t_node
	WHERE	id = newChild;
	
	-- unspecified in the Rec; check maximum / minimum number of children?
	
	SELECT	INTO
	newArchType	t_arch_type.id
	FROM	t_arch_type
	WHERE	t_arch_type.parent_node_type_id = thisNodeType AND
		t_arch_type.child_node_type_id = newChildType AND
		t_arch_type.specified_in_dom = TRUE;
		
	IF newArchType IS NULL
		THEN	SELECT	DOMException (3);
	END IF;
	
        -- WRONG_DOCUMENT_ERR: Raised if newChild was created from a different document than the one that created this node.
        IF	(SELECT	COUNT (DISTINCT COALESCE (ownerDocument, id)) > 1
        	FROM	t_Node
        	WHERE	id = thisNode OR
        		id = newChild)
        		
        	THEN	SELECT	DOMException (4);
        END IF;

        -- NO_MODIFICATION_ALLOWED_ERR: Raised if this node is readonly.
        -- not implemented yet; the Rec isn''t clear enough here

        -- NOT_FOUND_ERR: Raised if refChild is not a child of this node.
        IF	NOT EXISTS	(SELECT	id
        			FROM	t_arch
        			WHERE	parent_node_id = thisNode AND
        				child_node_id = refChild)
	        THEN	SELECT	DOMException (8);
        END IF;

        -- after all the error checking, start the actual work
        SELECT	INTO
        refChildOrdNo	child_ord_no
        FROM	t_arch
        WHERE	parent_node_id = thisNode AND
        	child_node_id = refChild;

        --BEGIN TRANSACTION
	
	-- It is important that DELETE precedes the UPDATE; consider moving sibilings around - a gap in numbering will occure
        DELETE
        FROM	t_arch
        WHERE	child_node_id = newChild AND
        	specified_in_dom = TRUE;
        	
        UPDATE	t_arch
        SET	child_ord_no = child_ord_no + 1
        WHERE	parent_node_id = thisNode AND
        	child_ord_no >= refChildOrdNo AND
        	specified_in_dom = TRUE;

        INSERT	INTO
        t_arch	(parent_node_id, child_node_id, parent_ord_no, child_ord_no, type_id, specified_in_dom)
        VALUES	(thisNode, newChild, 1, refChildOrdNo, newArchType, TRUE);

        --COMMIT TRANSACTION

	RETURN	newChild;
    END;	-- insertBefore
' LANGUAGE 'plpgsql';

-- Node                      replaceChild(in Node newChild,
-- in Node oldChild)
-- raises(DOMException);
CREATE	FUNCTION replaceChild (INT, INT, INT)	RETURNS INT
AS '
    DECLARE	thisNode	ALIAS FOR $1;
		newChild	ALIAS FOR $2;
    		oldChild	ALIAS FOR $3;
    		thisNodeType	SMALLINT;
    		newChildType	SMALLINT;
    		newArchType	INT := NULL;
    BEGIN
	-- HIERARCHY_REQUEST_ERR: Raised if this node is of a type that does not allow children of the type of the newChild node, or if the node to insert is one of this node''s ancestors.
	SELECT	INTO
	thisNodeType	NodeType
	FROM	t_node
	WHERE	id = thisNode;
	
	SELECT	INTO
	newChildType	NodeType
	FROM	t_node
	WHERE	id = newChild;
	
	-- unspecified in the Rec; check maximum / minimum number of children?
	
	SELECT	INTO
	newArchType	t_arch_type.id
	FROM	t_arch_type
	WHERE	t_arch_type.parent_node_type_id = thisNodeType AND
		t_arch_type.child_node_type_id = newChildType AND
		t_arch_type.specified_in_dom = TRUE;
		
	IF newArchType IS NULL
		THEN	SELECT	DOMException (3);
	END IF;
	
        -- WRONG_DOCUMENT_ERR: Raised if newChild was created from a different document than the one that created this node.
        IF	(SELECT	COUNT (DISTINCT COALESCE (ownerDocument, id)) > 1
        	FROM	t_Node
        	WHERE	id = thisNode OR
        		id = newChild)
        		
        	THEN	SELECT	DOMException (4);
        END IF;

        -- NO_MODIFICATION_ALLOWED_ERR: Raised if this node is readonly.
        -- not implemented yet; the Rec isn''t clear enough here

        -- NOT_FOUND_ERR: Raised if refChild is not a child of this node.
        IF	NOT EXISTS	(SELECT	id
        			FROM	t_arch
        			WHERE	parent_node_id = thisNode AND
        				child_node_id = refChild)
	        THEN	SELECT	DOMException (8);
        END IF;

        -- after all the error checking, start the actual work
        --BEGIN TRANSACTION
	
	DELETE
        FROM	t_arch
        WHERE	child_node_id = newChild AND
        	specified_in_dom = TRUE;
        	
        UPDATE	t_arch
        SET	child_node_id = newChild,
        	parent_ord_no = 1,
        	type_id = newArchType
        WHERE	parent_node_id = thisNode AND
        	child_node_id = oldChild AND
        	specified_in_dom = TRUE;

        --COMMIT TRANSACTION

	RETURN	oldChild;
    END;	-- replaceChild
' LANGUAGE 'plpgsql';


-- Node                      removeChild(in Node oldChild)
-- raises(DOMException);
CREATE	FUNCTION removeChild (INT, INT)	RETURNS INT
AS '
    DECLARE	thisNode	ALIAS FOR $1;
    		oldChild	ALIAS FOR $2;
    BEGIN

        -- NO_MODIFICATION_ALLOWED_ERR: Raised if this node is readonly.
        -- not implemented yet; the Rec isn''t clear enough here

        -- NOT_FOUND_ERR: Raised if oldChild is not a child of this node.
        IF	NOT EXISTS	(SELECT	id
        			FROM	t_arch
        			WHERE	parent_node_id = thisNode AND
        				child_node_id = oldChild)
        				-- consider checking and specified_in_dom
	        THEN	SELECT	DOMException (8);
        END IF;

        -- after all the error checking, start the actual work
        --BEGIN TRANSACTION
	
	DELETE
        FROM	t_arch
        WHERE	child_node_id = oldChild AND
        	parent_node_id = thisNode AND
        	specified_in_dom = TRUE;

        --COMMIT TRANSACTION

	RETURN	oldChild;
    END;	-- removeChild
' LANGUAGE 'plpgsql';


-- Node                      appendChild(in Node newChild)
-- raises(DOMException);
CREATE	FUNCTION appendChild (INT, INT)	RETURNS INT
AS '
    DECLARE	thisNode	ALIAS FOR $1;
		newChild	ALIAS FOR $2;
    		thisNodeType	SMALLINT;
    		newChildType	SMALLINT;
    		newArchType	INT := NULL;
    		maxNofChildren	INT;
    BEGIN
	-- HIERARCHY_REQUEST_ERR: Raised if this node is of a type that does not allow children of the type of the newChild node, or if the node to insert is one of this node''s ancestors.
	SELECT	INTO
	thisNodeType	NodeType
	FROM	t_node
	WHERE	id = thisNode;
	
	SELECT	INTO
	newChildType	NodeType
	FROM	t_node
	WHERE	id = newChild;
	
	-- unspecified in the Rec; check maximum / minimum number of children?
	
	SELECT	INTO
	newArchType, maxNofChildren
	t_arch_type.id, t_arch_type.max_nof_children
	FROM	t_arch_type
	WHERE	t_arch_type.parent_node_type_id = thisNodeType AND
		t_arch_type.child_node_type_id = newChildType AND
		t_arch_type.specified_in_dom = TRUE;
		
	IF newArchType IS NULL
		THEN	SELECT	DOMException (3);
	END IF;

	-- no outer joins in posrtgres
	IF      (SELECT COUNT (id) >= maxNofchildren
	        FROM    t_arch
	        WHERE   parent_node_id = thisNode AND
	                type_id = newArchType)
	        THEN    SELECT  DOMException (3);
	        -- consider an alternative exception
	END IF;
		
        -- WRONG_DOCUMENT_ERR: Raised if newChild was created from a different document than the one that created this node.
        IF	(SELECT	COUNT (DISTINCT COALESCE (ownerDocument, id)) > 1
        	FROM	t_Node
        	WHERE	id = thisNode OR
        		id = newChild)
        		
        	THEN	SELECT	DOMException (4);
        END IF;

        -- NO_MODIFICATION_ALLOWED_ERR: Raised if this node is readonly.
        -- not implemented yet; the Rec isn''t clear enough here

        --BEGIN TRANSACTION
	
	DELETE
        FROM	t_arch
        WHERE	child_node_id = newChild AND
        	specified_in_dom = TRUE;
        	
        INSERT	INTO
        t_arch	(parent_node_id, child_node_id, parent_ord_no, child_ord_no, type_id, specified_in_dom)
          (
        	SELECT	thisNode,
        		newChild,
        		1,
        		COALESCE( MAX(child_ord_no), 0) + 1,  -- either next ord_no, or 1, if this node has no children
        		newArchType,
        		TRUE
        	FROM	t_arch
        	WHERE	parent_node_id = thisNode AND
        		specified_in_dom = TRUE
          );

        --COMMIT TRANSACTION

	RETURN	newChild;
    END;	-- appendChild
' LANGUAGE 'plpgsql';


-- boolean                   hasChildNodes();
CREATE	FUNCTION hasChildNodes (INT)	RETURNS BOOLEAN
AS '
	SELECT	EXISTS	(SELECT	id
			FROM	t_arch
			WHERE	parent_node_id = $1 AND
				specified_in_dom = TRUE);
' LANGUAGE 'sql';

-- Node                      cloneNode(in boolean deep);


--------------------------------------------------------------------------
-- Interface NodeList
--------------------------------------------------------------------------
-- there are two NodeLists in DBDOM: ChildNodeList and DescendantNodeList
-- Node                      item(in unsigned long index);
CREATE FUNCTION childNodeItem (INT, INT) RETURNS INT
-- $1 - parent node
-- $2 - index
AS '
	select  child_node_id
	from    t_arch as t_main_arch
	where   t_main_arch.parent_node_id = $1 and
	        t_main_arch.specified_in_dom and
	        (select count (t_preceding_archs.id)
	        from    t_arch as t_preceding_archs
	        where   t_preceding_archs.parent_node_id = t_main_arch.parent_node_id and
	                t_preceding_archs.specified_in_dom and
	                t_preceding_archs.child_ord_no  < t_main_arch.child_ord_no) = $2;
' LANGUAGE 'sql';

-- readonly attribute  unsigned long        length;
CREATE FUNCTION getChildNodeListLength (INT) RETURNS INT
-- $1 - parent node
AS '
	SELECT  COUNT (child_node_id)
	FROM    t_arch
	WHERE   parent_node_id = $1 and
	        specified_in_dom;
' LANGUAGE 'sql';


--------------------------------------------------------------------------
-- Interface NamedNodeMap
--------------------------------------------------------------------------
-- a general assumption about NamedNodeMaps is that they only contain nodes
-- of one type. the node map definition is generalized to become the
-- collection of nodes that are children of a specified node and are of a
-- specified node type.

-- Node                      getNamedItem(in DOMString name);
-- Node                      setNamedItem(in Node arg)
-- raises(DOMException);
-- Node                      removeNamedItem(in DOMString name)
-- raises(DOMException);

-- Node                      item(in unsigned long index);
-- name clash with NodeList
CREATE FUNCTION namedMapItem (INT, SMALLINT, INT)	RETURNS INT
-- returns a child node of the specified type that is the index'th child
-- of this node. indexing is 0-based. no specific order is requied by the Rec;
-- the order used here is t_arch.id, or the physical order of records.
-- this ensures that attribues, which don't have a child_ord_no, are returned
-- in an order that is consistent from invocation to invocation
--
-- note that specified_in_dom is not checked here, as attributes are to be
-- returned
--
-- parameters:
-- $1    parent node
-- $2    type of nodes in the map
-- $3    index
AS '
	select  child_node_id
	from    t_arch as t_main_arch,
	        t_node as t_main_node
	where   t_main_arch.parent_node_id = $1 and
	        t_main_arch.child_node_id = t_main_node.id and
	        t_main_node.NodeType = $2 and
	        (select count (t_preceding_archs.id)
	        from    t_arch as t_preceding_archs,
	                t_node as t_preceding_nodes
	        where   t_preceding_archs.parent_node_id = t_main_arch.parent_node_id and
	                t_preceding_archs.child_node_id = t_preceding_nodes.id and
	                t_preceding_nodes.NodeType = t_main_node.NodeType and
	                t_preceding_archs.id  < t_main_arch.id) = $3;
' LANGUAGE 'sql';

-- readonly attribute  unsigned long        length;
-- name clash with NodeList
CREATE FUNCTION getNamedMapLength (INT, SMALLINT)	RETURNS INT
-- note that specified_in_dom is not checked here, as attributes are to be
-- returned
--
-- parameters:
-- $1    parent node
-- $2    type of nodes in the map
AS '
	SELECT  COUNT (t_arch.id)
	FROM    t_arch,
	        t_node
	WHERE   t_arch.child_node_id = t_node.id AND
	        t_arch.parent_node_id = $1 AND
	        t_node.NodeType = $2;
' LANGUAGE 'sql';

--------------------------------------------------------------------------
-- Interface CharacterData
--------------------------------------------------------------------------
-- attribute  DOMString            data;
-- // raises(DOMException) on setting
CREATE  FUNCTION setData (INT, VARCHAR)		RETURNS INT  -- OPAQUE
AS '
    DECLARE	thisNode	ALIAS FOR $1;
                argData         ALIAS FOR $2;
                nodeIsText	BOOLEAN;
    BEGIN
        -- this method doesn''t use nodeTypeMatches(INT), because it would take
        -- three consecutive SELECTs to get a result
        SELECT
        INTO    nodeIsText
                (NodeType = 3 OR NodeType = 4 OR NodeType = 8)
        FROM    t_Node
        WHERE   id = thisNode;

        -- INVALID_ACCESS_ERR
        IF NOT nodeIsText THEN
            SELECT DOMEXception(15);
        END IF;
        IF nodeIsText IS NULL THEN
            RAISE EXCEPTION ''Node doesn''''t exist'';
        END IF;

        -- INVALID_CHARACTER_ERR

        IF containsInvalidCharacters (argData) THEN
            SELECT DOMException(5);
        END IF;

        -- if everythig checks, go on

        UPDATE  t_CharacterData
        SET     data = argData
        WHERE   id = thisNode;

        RETURN NULL;
    END;
' LANGUAGE 'plpgsql';

-- // raises(DOMException) on retrieval
CREATE  FUNCTION getData (INT)		RETURNS VARCHAR
AS '
    DECLARE	thisNode	ALIAS FOR $1;
                nodeIsText	BOOLEAN;
                result		VARCHAR;
    BEGIN
        -- this method doesn''t use nodeTypeMatches(INT), because it would take
        -- three consecutive SELECTs to get a result
        SELECT
        INTO    nodeIsText
                (NodeType = 3 OR NodeType = 4 OR NodeType = 8)
        FROM    t_Node
        WHERE   id = thisNode;

        -- INVALID_ACCESS_ERR
        IF NOT nodeIsText THEN
            SELECT DOMEXception(15);
        END IF;
        IF nodeIsText IS NULL THEN
            RAISE EXCEPTION ''Node doesn''''t exist'';
        END IF;

        -- if everythig checks, go on

        SELECT
        INTO    result
                data
        FROM    t_CharacterData
        WHERE   id = thisNode;

        RETURN result;
    END;
' LANGUAGE 'plpgsql';

-- readonly attribute  unsigned long        length;
CREATE  FUNCTION getLength (INT)		RETURNS INT
AS '
    DECLARE	thisNode	ALIAS FOR $1;
                nodeIsText	BOOLEAN;
                result		INT;
    BEGIN
        -- this method doesn''t use nodeTypeMatches(INT), because it would take
        -- three consecutive SELECTs to get a result
        SELECT
        INTO    nodeIsText
                (NodeType = 3 OR NodeType = 4 OR NodeType = 8)
        FROM    t_Node
        WHERE   id = thisNode;

        -- INVALID_ACCESS_ERR
        IF NOT nodeIsText THEN
            SELECT DOMEXception(15);
        END IF;
        IF nodeIsText IS NULL THEN
            RAISE EXCEPTION ''Node doesn''''t exist'';
        END IF;

        -- if everythig checks, go on

        SELECT
        INTO    result
                char_length (data)
        FROM    t_CharacterData
        WHERE   id = thisNode;

        RETURN result;
    END;
' LANGUAGE 'plpgsql';

-- DOMString                 substringData(in unsigned long offset,
-- in unsigned long count)
-- raises(DOMException);
-- void                      appendData(in DOMString arg)
-- raises(DOMException);
CREATE FUNCTION	appendData (INT, VARCHAR) RETURNS BOOLEAN  -- ignorable
AS '
    UPDATE  t_CharacterData
    SET     data = data || $2
    WHERE   id = $1;

    SELECT  TRUE;

' LANGUAGE 'sql';

-- void                      insertData(in unsigned long offset,
-- in DOMString arg)
-- raises(DOMException);
-- void                      deleteData(in unsigned long offset,
-- in unsigned long count)
-- raises(DOMException);
-- void                      replaceData(in unsigned long offset,
-- in unsigned long count,
-- in DOMString arg)
-- raises(DOMException);


--------------------------------------------------------------------------
-- Interface Attr
--------------------------------------------------------------------------
-- readonly attribute  DOMString            name;
CREATE  FUNCTION getName (INT)		RETURNS VARCHAR
AS '
    DECLARE	thisNode	ALIAS FOR $1;
                nodeIsAttr	BOOLEAN;
                result		VARCHAR;
    BEGIN
        nodeIsAttr := nodeTypeMatches (thisNode, 2);
        -- INVALID_ACCESS_ERR
        IF NOT nodeIsAttr THEN
            SELECT DOMEXception(15);
        END IF;
        IF nodeIsAttr IS NULL THEN
            RAISE EXCEPTION ''Node doesn''''t exist'';
        END IF;

        -- if everythig checks, go on

        SELECT
        INTO    result
                name
        FROM    t_Attr,
                t_attr_name
        WHERE   t_Attr.id = thisNode AND
                t_Attr.name_id = t_attr_name.id;

        RETURN  result;
    END;
' LANGUAGE 'plpgsql';

-- readonly attribute  boolean              specified;
CREATE  FUNCTION getSpecified (INT)		RETURNS BOOLEAN
AS '
    DECLARE	thisNode	ALIAS FOR $1;
                nodeIsAttr	BOOLEAN;
                result		BOOLEAN;
    BEGIN
        nodeIsAttr := nodeTypeMatches (thisNode, 2);
        -- INVALID_ACCESS_ERR
        IF NOT nodeIsAttr THEN
            SELECT DOMEXception(15);
        END IF;
        IF nodeIsAttr IS NULL THEN
            RAISE EXCEPTION ''Node doesn''''t exist'';
        END IF;

        -- if everythig checks, go on

        SELECT
        INTO    result
                specified
        FROM    t_Attr
        WHERE   id = thisNode;

        RETURN  result;
    END;
' LANGUAGE 'plpgsql';


-- attribute  DOMString            value;
CREATE  FUNCTION getValue (INT)		RETURNS VARCHAR
-- this function works with the current schema, which is broken in its
-- treatment of attribute values; what should be returned is concatinated
-- values of all text children and the values of entities
AS '
    DECLARE	thisNode	ALIAS FOR $1;
                nodeIsAttr	BOOLEAN;
                result		VARCHAR;
    BEGIN
        nodeIsAttr := nodeTypeMatches (thisNode, 2);
        -- INVALID_ACCESS_ERR
        IF NOT nodeIsAttr THEN
            SELECT DOMEXception(15);
        END IF;
        IF nodeIsAttr IS NULL THEN
            RAISE EXCEPTION ''Node doesn''''t exist'';
        END IF;

        -- if everythig checks, go on

        SELECT
        INTO    result
                value
        FROM    t_Attr
        WHERE   id = thisNode;

        RETURN  COALESCE (result, '''');  -- value may be empty, but not null
    END;
' LANGUAGE 'plpgsql';


--------------------------------------------------------------------------
-- Interface Element
--------------------------------------------------------------------------
-- readonly attribute  DOMString            tagName;
CREATE	FUNCTION getTagName (INT)	RETURNS VARCHAR
AS '
    DECLARE	thisNode	ALIAS FOR $1;
                nodeIsElement	BOOLEAN;
                result		VARCHAR;
    BEGIN
        nodeIsElement := nodeTypeMatches (thisNode, 1);
        IF NOT nodeIsElement THEN
            SELECT DOMEXception(15);
        END IF;
        IF nodeIsElement IS NULL THEN
            RAISE EXCEPTION ''Node doesn''''t exist'';
        END IF;

        -- if everythig checks, go on

        SELECT
        INTO    result
                name
        FROM    t_Element,
                t_tag
        WHERE   t_Element.id = $1 AND
                t_Element.tag_id = t_tag.id;

        RETURN  result;
    END;
' LANGUAGE 'plpgsql';

-- DOMString                 getAttribute(in DOMString name);
CREATE	FUNCTION getAttribute (INT, VARCHAR)	RETURNS VARCHAR
AS '
    DECLARE	thisNode	ALIAS FOR $1;
                attrName	ALIAS FOR $2;
                nodeIsElement	BOOLEAN;
                result		VARCHAR;
    BEGIN
        nodeIsElement := nodeTypeMatches (thisNode, 1);
        IF NOT nodeIsElement THEN
            SELECT DOMEXception(15);
        END IF;
        IF nodeIsElement IS NULL THEN
            RAISE EXCEPTION ''Node doesn''''t exist'';
        END IF;

        -- if everythig checks, go on

        SELECT
        INTO    result
                t_Attr.value
        FROM    t_arch,
                t_Attr,
                t_attr_name
        WHERE   t_arch.parent_node_id = thisNode AND
                t_arch.child_node_id = t_Attr.id AND
                t_attr_name.name = attrName;

        RETURN  COALESCE (result, '''');
    END;
' LANGUAGE 'plpgsql';

-- void                      setAttribute(in DOMString name,
-- in DOMString value)
-- raises(DOMException);
CREATE	FUNCTION setAttribute (INT, VARCHAR, VARCHAR)	RETURNS BOOLEAN -- ignorable
AS '
    DECLARE	thisNode	ALIAS FOR $1;
    		attrName	ALIAS FOR $2;
    		attrValue	ALIAS FOR $3;
    		currAttr	INT;	-- current Attribute of thisNode that
                                        -- has the same attrName, if
                                        -- one exists
                newAttr         INT;    -- new Attribute, if one is created
    BEGIN
        -- see if thisNode has an attribute with the specified name
        SELECT
        INTO    currAttr
                child_node_id
        FROM    t_arch,
                t_Attr,
                t_attr_name
        WHERE   parent_node_id = thisNode AND
                child_node_id = t_Attr.id AND
                t_Attr.name_id = t_attr_name.id AND
                t_attr_name.name = attrName;

        IF currAttr IS NOT NULL THEN
            -- this node has an attribute with the specified name
            -- this uses the broken representation of Attr.value;
            -- fix when Attr.value is fixed
            UPDATE  t_Attr
            SET     value = attrValue
            WHERE   id = currAttr;

        ELSE
            -- thisNode doesn''t have an attribute with this name
            -- create new attribute node
            newAttr := createAttribute (getOwnerDocument(thisNode), attrName);
            UPDATE  t_Attr
            SET     value = attrValue
            WHERE   id = newAttr;

            -- create new arch
            INSERT  INTO
            t_arch (parent_node_id,
                    child_node_id,
                    type_id,
                    parent_ord_no,
                    child_ord_no,
                    specified_in_dom)
                   (SELECT  thisNode,
                            newAttr,
                            id,
                            1,        -- attributes are unordered
                            1,        -- attributes are unordered
                            FALSE
                    FROM    t_arch_type
                    WHERE   parent_node_type_id = 1 AND
                            child_node_type_id = 2);
        END IF;

        RETURN NULL;
    END;
' LANGUAGE 'plpgsql';

-- void                      removeAttribute(in DOMString name)
-- raises(DOMException);
-- Attr                      getAttributeNode(in DOMString name);
CREATE	FUNCTION getAttributeNode (INT, VARCHAR)	RETURNS INT
AS '
    DECLARE	thisNode	ALIAS FOR $1;
                attrName	ALIAS FOR $2;
                nodeIsElement	BOOLEAN;
                result		INT;
    BEGIN
        nodeIsElement := nodeTypeMatches (thisNode, 1);
        IF NOT nodeIsElement THEN
            SELECT DOMEXception(15);
        END IF;
        IF nodeIsElement IS NULL THEN
            RAISE EXCEPTION ''Node doesn''''t exist'';
        END IF;

        -- if everythig checks, go on

        SELECT
        INTO    result
                t_Attr.id
        FROM    t_arch,
                t_Attr,
                t_attr_name
        WHERE   t_arch.parent_node_id = thisNode AND
                t_arch.child_node_id = t_Attr.id AND
                t_attr_name.name = attrName;

        RETURN  result;
    END;
' LANGUAGE 'plpgsql';

-- Attr                      setAttributeNode(in Attr newAttr)
-- raises(DOMException);
CREATE	FUNCTION setAttributeNode (INT, INT)	RETURNS INT
AS '
    DECLARE	thisNode	ALIAS FOR $1;
                newAttr		ALIAS FOR $2;
                nodeIsElement	BOOLEAN;
                nodeIsAttr	BOOLEAN;
                currAttr        INT;	-- current Attribute of thisNode that
                                        -- has the same name as newAttr, if
                                        -- one exists
    BEGIN
        nodeIsElement := nodeTypeMatches (thisNode, 1);
        --INVALID_ACCESS_ERR
        IF NOT nodeIsElement THEN
            SELECT DOMEXception(15);
        END IF;
        IF nodeIsElement IS NULL THEN
            RAISE EXCEPTION ''Node doesn''''t exist'';
        END IF;

        nodeIsAttr := nodeTypeMatches (newAttr, 2);
        -- INVALID_ACCESS_ERR
        IF NOT nodeIsAttr THEN
            SELECT DOMEXception(15);
        END IF;
        IF nodeIsAttr IS NULL THEN
            RAISE EXCEPTION ''Node doesn''''t exist'';
        END IF;

        -- WRONG_DOCUMENT_ERR
        IF NOT (
            SELECT  element.ownerDocument = attr.ownerDocument
            FROM    t_Node as element,
                    t_Node as attr
            WHERE   element.id = thisNode AND
                    attr.id = newAttr) THEN
            SELECT DOMException(4);
        END IF;

        -- INUSE_ATTRIBUTE_ERR
        IF (SELECT  parent_node_id
            FROM    t_arch
            WHERE   parent_node_id != thisNode AND
                    child_node_id = newAttr) IS NOT NULL THEN
            SELECT DOMException(10);
        END IF;

        -- if everythig checks, go on

        -- see if thisNode has an attribute with newAttr''s name
        SELECT
        INTO    currAttr
                child_node_id
        FROM    t_arch
        WHERE   parent_node_id = thisNode AND
                child_node_id IN (
                SELECT  id
                FROM    t_Attr
                WHERE   name_id = (
                        SELECT  name_id
                        FROM    t_Attr
                        WHERE   id = newAttr
                )
        );

        IF currAttr = newAttr THEN
            -- nothing to do - newAttr is already an attribute of thisNode
            RETURN newAttr;
        END IF;

        IF currAttr IS NULL THEN
            -- thisNode doesn''t have an attribute with this name
            -- newAttr has no parent - INUSE_ERROR would have been thrown
            -- create new arch
            INSERT  INTO
            t_arch (parent_node_id,
                    child_node_id,
                    type_id,
                    parent_ord_no,
                    child_ord_no,
                    specified_in_dom)
                   (SELECT  thisNode,
                            newAttr,
                            id,
                            1,        -- attributes are unordered
                            1,        -- attributes are unordered
                            FALSE
                    FROM    t_arch_type
                    WHERE   parent_node_type_id = 1 AND
                            child_node_type_id = 2);
            RETURN newAttr;
        END IF;

        -- if none of the above is true, thisNode has an attribute with this name
        -- replace the attribute
        -- consider garbage-collecting currAttr - it is now completely unreachable
        UPDATE  t_arch
        SET     child_node_id = newAttr
        WHERE   parent_node_id = thisNode AND
                child_node_id = currAttr;

        RETURN  newAttr;
    END;  -- setAttributeNode
' LANGUAGE 'plpgsql';

-- Attr                      removeAttributeNode(in Attr oldAttr)
-- raises(DOMException);
-- NodeList                  getElementsByTagName(in DOMString name);
-- void                      normalize();


--------------------------------------------------------------------------
-- Interface Text
--------------------------------------------------------------------------
-- Text                      splitText(in unsigned long offset)
-- raises(DOMException);


--------------------------------------------------------------------------
-- Interface Comment
--------------------------------------------------------------------------


--------------------------------------------------------------------------
-- Interface CDATASection
--------------------------------------------------------------------------


--------------------------------------------------------------------------
-- Interface DocumentType
--------------------------------------------------------------------------
-- readonly attribute  DOMString            name;
-- readonly attribute  NamedNodeMap         entities;
-- readonly attribute  NamedNodeMap         notations;


--------------------------------------------------------------------------
-- Interface Notation
--------------------------------------------------------------------------
-- readonly attribute  DOMString            publicId;
-- readonly attribute  DOMString            systemId;


--------------------------------------------------------------------------
-- Interface Entity
--------------------------------------------------------------------------
-- readonly attribute  DOMString            publicId;
-- readonly attribute  DOMString            systemId;
-- readonly attribute  DOMString            notationName;


--------------------------------------------------------------------------
-- Interface EntityReference
--------------------------------------------------------------------------


--------------------------------------------------------------------------
-- Interface ProcessingInstruction
--------------------------------------------------------------------------
-- readonly attribute  DOMString            target;
CREATE FUNCTION getTarget	(INT)	RETURNS VARCHAR
AS '
        SELECT  target
        FROM    t_ProcessingInstruction
        WHERE   id = $1;
' LANGUAGE 'sql';

-- attribute  DOMString            data;
-- // raises(DOMException) on setting
-- this attribute is renamed pIdata to avoid name clash with interface CharacterData
CREATE FUNCTION getPiData	(INT)	RETURNS VARCHAR
AS '
        SELECT  data
        FROM    t_ProcessingInstruction
        WHERE   id = $1;
' LANGUAGE 'sql';


CREATE FUNCTION setPiData	(INT, VARCHAR)	RETURNS BOOLEAN --ignorable
AS '
        UPDATE  t_ProcessingInstruction
        SET     data = $2
        WHERE   id = $1;

        SELECT  TRUE;
' LANGUAGE 'sql';
